"""
This submodule contains Optimizer classes to learn an MHN from mutation data.
"""
# author(s): Stefan Vocht, Y. Linda Hu, Rudolf Schill, Andreas Lösch

from __future__ import annotations

import abc
import warnings
from enum import Enum
from typing import Callable

import numpy as np
import pandas as pd
from tqdm.auto import trange

from mhn.training import (likelihood_cmhn, likelihood_omhn, penalties_cmhn,
                          penalties_omhn)
from mhn.training import regularized_optimization as reg_optim

from . import model
from .training.likelihood_cmhn import CUDA_AVAILABLE, CUDAError, cuda_available
from .training.state_containers import StateContainer, create_indep_model


class Device(Enum):
    """
    Enum of device types.

    Attributes:
        AUTO (int): Automatically selects the device based on the number of active events in each sample.
        CPU (int): Executes all computations on the CPU.
        GPU (int): Executes score and gradient computations on the GPU.
    """

    AUTO, CPU, GPU = range(3)


class Penalty(Enum):
    """
    Enumeration of penalty functions.

    Attributes:
        L1 (int): Applies L1 regularization during training.
        L2 (int): Applies L2 regularization during training.
        SYM_SPARSE (int): Induces sparsity and soft symmetry (see Schill et al., 2024).
    """

    L1, L2, SYM_SPARSE = range(3)


class _Optimizer(abc.ABC):
    """
    Abstract Optimizer class serving as a base for other optimizer classes.

    Attributes:
        Device (Device): Enum reference for device types.
        Penalty (Penalty): Enum reference for penalty types.
    """

    # Reference to the external enum (re-export), makes separate import of Enums unnecessary
    Device = Device
    Penalty = Penalty

    def __init__(self):
        self._data = None
        self._bin_datamatrix = None
        self._result = None
        self._events = None

        self._init_theta = None
        self.__custom_callback = None

        self.__backup_steps = -1
        self.__backup_filename = None
        self.__backup_always_new_file = False
        self.__backup_current_step = None

        self._gradient_and_score_func = None
        self._regularized_score_func_builder = lambda grad_score_func: \
            penalties_cmhn.build_regularized_score_func(
                grad_score_func, penalties_cmhn.l1)
        self._regularized_gradient_func_builder = lambda grad_score_func: \
            penalties_cmhn.build_regularized_gradient_func(
                grad_score_func, penalties_cmhn.l1_)

        self._penalty = (penalties_cmhn.l1, penalties_cmhn.l1_)

        self._OutputMHNClass = model.cMHN

    def set_init_theta(self, init: np.ndarray | None) -> _Optimizer:
        """
        Sets the initial theta matrix for learning a new MHN.

        Args:
            init (np.ndarray | None): Initial theta matrix in logarithmic form. If None, uses an independence model where the baseline hazard Theta_ii
                                      of each event is set to its empirical odds and the hazard ratios (off-diagonal entries) are set to exactly 1.

        Returns:
            _Optimizer: The optimizer instance.
        """
        self._init_theta = init
        return self

    def get_data_properties(self) -> dict:
        """
        Retrieves properties of the loaded training data.

        Returns:
            dict: A dictionary with information about the training data, including sample and event statistics.
        """
        if self._bin_datamatrix is None:
            return {}

        total_event_occurrence = np.sum(self._bin_datamatrix, axis=0)
        event_frequencies = total_event_occurrence / \
            self._bin_datamatrix.shape[0]
        event_dataframe = pd.DataFrame.from_dict({
            "Total": total_event_occurrence,
            "Frequency": event_frequencies
        })
        if self._events is not None:
            event_dataframe.index = self._events

        total_events_per_sample = np.sum(self._bin_datamatrix, axis=1)
        return {
            'samples': self._bin_datamatrix.shape[0],
            'events': self._bin_datamatrix.shape[1],
            'occurred events per sample': {
                'mean': np.mean(total_events_per_sample),
                'median': np.median(total_events_per_sample),
                'max': np.max(total_events_per_sample),
                'min': np.min(total_events_per_sample)
            },
            'event statistics': event_dataframe
        }

    def set_callback_func(self, callback=None) -> _Optimizer:
        """
        Sets a callback function to be invoked after each iteration of the BFGS algorithm.

        Args:
            callback (Callable): A function that takes a single argument (theta matrix computed in the last iteration). Defaults to None.

        Raises:
            ValueError: If the provided callback is not callable.

        Returns:
            _Optimizer: The optimizer instance.
        """
        if callback is not None and not callable(callback):
            raise ValueError("callback has to be a function!")
        self.__custom_callback = callback
        return self

    def save_progress(self, steps: int = -1, always_new_file: bool = False, filename: str = 'theta_backup.npy') -> _Optimizer:
        """
        Configures periodic saving of training progress.

        Args:
            steps (int): Number of training iterations between progress saves. Defaults to -1 (disabled).
            always_new_file (bool): Whether to save each backup as a new file. Defaults to False.
            filename (str): Name of the backup file. Defaults to 'theta_backup.npy'.

        Returns:
            _Optimizer: The optimizer instance.
        """
        self.__backup_steps = steps
        self.__backup_always_new_file = always_new_file
        self.__backup_filename = filename
        return self

    def __total_callback_func(self, theta: np.ndarray):
        if self.__custom_callback is not None:
            self.__custom_callback(theta)

        if self.__backup_steps > 0:
            self.__create_backup(theta)

    def __create_backup(self, theta: np.ndarray):
        self.__backup_current_step += 1
        if (self.__backup_current_step % self.__backup_steps) == 0:
            filename = self.__backup_filename
            if self.__backup_always_new_file:
                try:
                    idx = filename.index(".")
                    filename = filename[:idx] + \
                        f"_{self.__backup_current_step}" + filename[idx:]
                except ValueError:  # str.index raises ValueError if no "." is present in the filename
                    filename += f"_{self.__backup_current_step}.npy"
            with open(filename, 'wb') as f:
                np.save(f, theta)

    def train(self, lam: float = None, maxit: int = 5000, trace: bool = False,
              reltol: float = 1e-7, round_result: bool = True) -> model.cMHN:
        """
        Trains a new MHN model using the loaded data.

        Args:
            lam (float, optional): Regularization parameter. Defaults to 1/(number of samples).
            maxit (int): Maximum number of training iterations. Defaults to 5000.
            trace (bool): Whether to print convergence messages. Defaults to False.
            reltol (float): Gradient norm tolerance for termination. Defaults to 1e-7. (see "gtol" in scipy.optimize.minimize)
            round_result (bool): Whether to round the result to two decimal places. Defaults to True.

        Returns:
            model.cMHN: The trained MHN model.

        Raises:
            ValueError: If no data has been loaded.
        """
        if self._data is None:
            raise ValueError("You have to load data before training!")

        if lam is None:
            lam = 1 / self._data.get_data_shape()[0]
            
        self._result = None
        self.__backup_current_step = 0

        if self.__custom_callback is None and self.__backup_steps < 1:
            callback_func = None
        else:
            callback_func = self.__total_callback_func

        score_func = self._regularized_score_func_builder(
            self._gradient_and_score_func)
        gradient_func = self._regularized_gradient_func_builder(
            self._gradient_and_score_func)

        result = reg_optim.learn_mhn(self._data, score_func, gradient_func, self._init_theta, lam, maxit, trace, reltol,
                                     round_result, callback_func)

        self.__backup_current_step = None

        self._result = self._OutputMHNClass(
            log_theta=result.x,
            events=self._events,
            meta={
                "lambda": lam,
                "init": self._init_theta,
                "maxit": maxit,
                "reltol": reltol,
                "score": result.fun,
                "message": result.message,
                "status": result.status,
                "nit": result.nit
            })

        return self._result

    @property
    def result(self) -> model.cMHN:
        """
        Property for retrieving the training result.

        Returns:
            model.cMHN: The resulting MHN model after training.
        """
        return self._result

    @property
    @abc.abstractmethod
    def training_data(self):
        """
        This property returns all the data given to this optimizer to train a new cMHN.
        """
        pass

    @staticmethod
    def _preprocess_binary_matrix(data_matrix: np.ndarray) -> np.ndarray:
        """
        Preprocesses a binary data matrix to ensure the correct format.

        Args:
            data_matrix (np.ndarray): Input binary matrix.

        Returns:
            np.ndarray: Preprocessed data matrix with dtype set to np.int32.

        Raises:
            ValueError: If the matrix is not two-dimensional or contains values other than 0 and 1.
        """
        if len(data_matrix.shape) != 2:
            raise ValueError("The given data matrix must be two-dimensional")
        # StateContainer only accepts numpy arrays with dtype=np.int32
        if data_matrix.dtype != np.int32:
            data_matrix = data_matrix.astype(dtype=np.int32)
            warnings.warn(
                "The dtype of the given data matrix is changed to np.int32")
        if not set(data_matrix.flatten()).issubset({0, 1}):
            raise ValueError("The data matrix must only contain 0s and 1s")

        return data_matrix

    @abc.abstractmethod
    def set_device(self, device: Device) -> _Optimizer:
        """
        Sets the computational device for training.

        You have three options:
            Device.AUTO: (default) automatically select the device that is likely to match the data
            Device.CPU:  use the CPU implementations to compute the scores and gradients
            Device.GPU:  use the GPU/CUDA implementations to compute the scores and gradients

        The Device enum is part of this optimizer class.

        Args:
            device (Device): The device to use (AUTO, CPU, or GPU).

        Returns:
            _Optimizer: The optimizer instance.

        Raises:
            ValueError: If the given device is not an instance of Device.
        """
        if not isinstance(device, Device):
            raise ValueError(
                f"The given device is not an instance of {Device}")

        return self

    def set_penalty(
        self,
        penalty: (
            Penalty
            | tuple[
                Callable[[np.ndarray], float],
                Callable[[np.ndarray], np.ndarray],
            ]
        ),
    ) -> _Optimizer:
        """
        Sets the penalty type for training.

        You have four options:
            Penalty.L1:                 (default) uses the L1 penalty as
                regularization
            Penalty.L2:                 uses the L2 penalty as
                regularization
            Penalty.SYM_SPARSE:         uses a penalty which induces
                sparsity and soft symmetry
            (penalty, penalty_deriv):   a custom penalty function and
                its derivative. They must both take the log_theta matrix
                in the form of a numpy array as input and return a float
                and a numpy array, respectively.

        The Penalty enum is part of this optimizer class.

        Args:
            penalty (Penalty | tuple[Callable[[np.ndarray], float], Callable[[np.ndarray], np.ndarray]]): The penalty to use (L1, L2, SYM_SPARSE).

        Returns:
            _Optimizer: The optimizer instance.

        Raises:
            ValueError: If the given penalty is not an instance of
            Penalty or a tuple of two functions.
        """
        if not isinstance(penalty, oMHNOptimizer.Penalty) and (
            not isinstance(penalty, tuple)
            or len(penalty) != 2
            or not callable(penalty[0])
            or not callable(penalty[1])
        ):
            raise ValueError(
                "The given penalty must either be an instance of _Optimizer.Penalty or a tuple of two functions."
            )
        
        if isinstance(penalty, oMHNOptimizer.Penalty):
            penalty_score, penalty_gradient = {
                Penalty.L1: (penalties_cmhn.l1, penalties_cmhn.l1_),
                Penalty.L2: (penalties_cmhn.l2, penalties_cmhn.l2_),
                Penalty.SYM_SPARSE: (
                    penalties_cmhn.sym_sparse,
                    penalties_cmhn.sym_sparse_deriv,
                ),
            }[penalty]
        else:
            penalty_score, penalty_gradient = penalty
        
        self._regularized_score_func_builder = lambda grad_score_func: penalties_cmhn.build_regularized_score_func(
            grad_score_func, penalty_score
        )
        self._regularized_gradient_func_builder = lambda grad_score_func: penalties_cmhn.build_regularized_gradient_func(
            grad_score_func, penalty_gradient
        )

        self._penalty = (penalty_score, penalty_gradient)

        return self

    @property
    def penalty(self):
        return self._penalty

    def __getstate__(self):
        state = self.__dict__.copy()
        state.pop("_data")
        state.pop("_regularized_score_func_builder")
        state.pop("_regularized_gradient_func_builder")
        
        return state
    
    def __setstate__(self, state):
        self.__dict__.update(state)
        self._data = StateContainer(state["_bin_datamatrix"])

class cMHNOptimizer(_Optimizer):
    """
    Optimizes an cMHN for given cross-sectional data.
    """

    def __init__(self):
        super().__init__()
        self._gradient_and_score_func = likelihood_cmhn.gradient_and_score

    def load_data_matrix(self, data_matrix: np.ndarray | pd.DataFrame):
        """
        Loads mutation data stored in a numpy array or pandas DataFrame.

        Args:
            data_matrix (np.ndarray | pd.DataFrame): Data matrix where rows represent samples and columns represent genes.
                                                     Mutations of genes are represented by 1s, intact genes by 0s.

        Returns:
            cMHNOptimizer: This optimizer object.
        """
        if isinstance(data_matrix, pd.DataFrame):
            self._events = data_matrix.columns.to_list()
            data_matrix = np.array(data_matrix, dtype=np.int32)
        else:
            self._events = None
        data_matrix = self._preprocess_binary_matrix(data_matrix)
        self._data = StateContainer(data_matrix)
        self._bin_datamatrix = data_matrix
        return self

    def load_data_from_csv(self, src: str, delimiter: str = ',', **kwargs):
        """
        Load mutation data from a CSV file. The rows have to represent samples and the columns represent genes.
        Mutations of genes are represented by 1s, intact genes are represented by 0s.

        Args:
            src (str): Path to the CSV file.
            delimiter (str, optional): Delimiter used in the CSV file (default is ',').
            kwargs: Additional keyword arguments passed to pandas' read_csv() function.

        Returns:
            cMHNOptimizer: This optimizer object.
        """
        df = pd.read_csv(src, delimiter=delimiter, **kwargs)
        self.load_data_matrix(df)
        return self

    def lambda_from_cv(self, lambda_min: float | None = None, lambda_max: float | None = None,
                       steps: int = 9, nfolds: int = 5, lambda_vector: np.ndarray | None = None,
                       show_progressbar: bool = False, return_lambda_scores: bool = False, pick_1se: bool = True,
                       ) -> float | tuple[float, pd.DataFrame]:
        """
        Finds the best value for lambda according to either the maximal average test set likelihood or the
        "one-standard-error-rule" through n-fold cross-validation.

        You can specify the lambda values that should be tested in cross-validation by setting the lambda_vector
        parameter accordingly.

        Alternatively, you can specify the minimum, maximum and step size for potential lambda values. This method
        will then create a range of possible lambdas with logarithmic grid-spacing, e.g. (0.0001, 0.0010, 0.0100, 0.1000)
        for lambda_min=0.0001, lambda_max=0.1 and steps=4.

        If you set neither lambda_vector nor lambda_min and lambda_max, the default range (0.1/#datasamples, 10/#datasamples)
        will be used.

        By default, the largest lambda that performed within one standard error of the best-performing lambda is returned as the
        preferred lambda ("one-standard-error-rule"). When setting pick_1se=False, the function will simply return the best-performing lambda instead.

        Use np.random.seed() to make results reproducible.

        Args:
            lambda_min (float, optional): Minimum lambda value to test. Will be ignored if lambda_vector is set.
            lambda_max (float, optional): Maximum lambda value to test. Will be ignored if lambda_vector is set.
            steps (int, optional): Number of steps between lambda_min and lambda_max. Defaults to 9. Will be ignored if lambda_vector is set.
            nfolds (int, optional): Number of folds for cross-validation. Defaults to 5.
            lambda_vector (np.ndarray, optional): Specific lambda values to test.
            show_progressbar (bool, optional): Whether to show a progress bar during cross-validation. Defaults to False.
            return_lambda_scores (bool, optional): Whether to return lambda scores along with the best lambda. Defaults to False.
            pick_1se (bool, optional): If True (default), applies the one-standard-error-rule to pick the returned lambda value. If False,
                                       returns the best-performing lambda.

        Returns:
            float | tuple[float, pd.DataFrame]: Best lambda value, or, if return_lambda_scores is set to True, a tuple with the best lambda and
                                                a DataFrame containing the mean scores for each lambda.
        """
        if self._bin_datamatrix is None:
            raise ValueError(
                "You have to load data before you start cross-validation")

        if lambda_min is None and lambda_max is not None or lambda_min is not None and lambda_max is None:
            raise ValueError("You have to set both lambda_min and lambda_max, if you want to use them.")

        if lambda_min is None and lambda_max is None:
            # the default lambda value used in train() if lambda is not set
            default_lambda = 1 / self._data.get_data_shape()[0]
            lambda_min = 0.1 * default_lambda
            lambda_max = 10 * default_lambda

        if lambda_vector is None:
            # create a range of possible lambdas with logarithmic grid-spacing
            # e.g. (0.0001,0.0010,0.0100,0.1000) for 4 steps
            lambda_path: np.ndarray = np.exp(np.linspace(
                np.log(lambda_min + 1e-10), np.log(lambda_max + 1e-10), steps))
        else:
            lambda_path = lambda_vector
            steps = lambda_vector.size

        # shuffle the dataset and cut it into n folds
        shuffled_data = self._bin_datamatrix.copy()
        np.random.shuffle(shuffled_data)
        folds = np.arange(self._bin_datamatrix.shape[0]) % nfolds

        # store the scores for each fold in rows and each lambda in columns
        scores = np.zeros((nfolds, steps))

        # use self.__class__ to make this method also usable for subclasses
        opt = self.__class__()

        # make sure that the same score, gradient and regularization functions are used
        opt._gradient_and_score_func = self._gradient_and_score_func
        opt._regularized_score_func_builder = self._regularized_score_func_builder
        opt._regularized_gradient_func_builder = self._regularized_gradient_func_builder

        disable_progressbar = not show_progressbar

        for j in trange(nfolds, desc="Cross-Validation Folds", position=0, disable=disable_progressbar):
            # designate one of folds as test set and the others as training set
            test_data = shuffled_data[np.where(folds == j)]
            test_data_container = StateContainer(test_data)
            train_data = shuffled_data[np.where(folds != j)]
            opt.load_data_matrix(train_data)

            for i in trange(steps, desc="Lambda Evaluation", position=1, leave=False, disable=disable_progressbar):
                opt.train(lam=lambda_path[i].item())
                theta = opt.result.log_theta
                # make sure that events have not a rate of zero, which can become
                # a problem if the event is present in the test set, in which case
                # this would lead to score = -inf
                minimum_base_rate = -np.log(train_data.shape[0] + 1)  # log(1 / (#samples + 1))
                theta[theta < minimum_base_rate] = minimum_base_rate
                scores[j, i] = self._gradient_and_score_func(
                    theta, test_data_container)[1]

        # find the best performing lambda with the highest average score over folds
        score_means = np.sum(scores, axis=0) / nfolds
        best_lambda_idx = np.argmax(score_means)

        if pick_1se:
            # choose the actual lambda according to the "one standard error rule"
            standard_error = np.std(scores[:, best_lambda_idx]) / np.sqrt(nfolds)
            threshold = np.max(score_means) - standard_error
            chosen_lambda_idx = np.max(np.argwhere(score_means >= threshold))
        else:
            # simply choose the best-performing lambda
            chosen_lambda_idx = best_lambda_idx

        chosen_lambda = lambda_path[chosen_lambda_idx].item()

        if not lambda_path.min() < chosen_lambda < lambda_path.max():
            warnings.warn(
                "Optimal lambda is at a limit (min/max) of the given search range. Consider re-running with adjusted search range.")

        if return_lambda_scores:
            score_dataframe = pd.DataFrame.from_dict({
                "Lambda Value": lambda_path,
                "Mean Score": score_means,
                "Standard Error": np.std(scores, axis=0) / np.sqrt(nfolds)
            })
            return chosen_lambda, score_dataframe

        return chosen_lambda

    def set_device(self, device: Device) -> cMHNOptimizer:
        """
        Sets the computational device for training.

        You have three options:
            Device.AUTO: (default) automatically select the device that is likely to match the data
            Device.CPU:  use the CPU implementations to compute the scores and gradients
            Device.GPU:  use the GPU/CUDA implementations to compute the scores and gradients

        The Device enum is part of this optimizer class.

        Args:
            device (Device): The device to use (AUTO, CPU, or GPU).

        Returns:
            cMHNOptimizer: The optimizer instance.

        Raises:
            ValueError: If the given device is not an instance of Device.
        """
        super().set_device(device)
        if device == Device.GPU:
            if cuda_available() != CUDA_AVAILABLE:
                raise CUDAError(cuda_available())

            self._gradient_and_score_func = likelihood_cmhn.cuda_gradient_and_score
        else:
            self._gradient_and_score_func = {
                Device.AUTO: likelihood_cmhn.gradient_and_score,
                Device.CPU: likelihood_cmhn.cpu_gradient_and_score
            }[device]
        return self

    @property
    def training_data(self) -> np.ndarray:
        """
        Returns the data used to train the cMHN model.

        Returns:
            np.ndarray: The data matrix used for training.
        """
        return self._bin_datamatrix


class oMHNOptimizer(cMHNOptimizer):
    """
    Optimizer for the oMHN model.
    """

    def __init__(self):
        super().__init__()
        self._gradient_and_score_func = likelihood_omhn.gradient_and_score
        self._regularized_score_func_builder = lambda grad_score_func: \
            penalties_omhn.build_regularized_score_func(
                grad_score_func, penalties_omhn.l1)
        self._regularized_gradient_func_builder = lambda grad_score_func: \
            penalties_omhn.build_regularized_gradient_func(
                grad_score_func, penalties_omhn.l1_)
        self._OutputMHNClass = model.oMHN

        self._penalty = (penalties_omhn.l1, penalties_omhn.l1_)


    def train(self, lam: float = None, maxit: int = 5000, trace: bool = False,
              reltol: float = 1e-7, round_result: bool = True) -> model.oMHN:
        """
        Trains a new oMHN model using the loaded data.

        Args:
            lam (float, optional): Regularization parameter. Defaults to 1/(number of samples).
            maxit (int): Maximum number of training iterations. Defaults to 5000.
            trace (bool): Whether to print convergence messages. Defaults to False.
            reltol (float): Gradient norm tolerance for termination. Defaults to 1e-7. (see "gtol" in scipy.optimize.minimize)
            round_result (bool): Whether to round the result to two decimal places. Defaults to True.

        Returns:
            model.oMHN: The trained MHN model.

        Raises:
            ValueError: If no data has been loaded.
        """
        if self._data is None:
            raise ValueError("You have to load data before training!")

        undo_init_theta = False
        if self._init_theta is None:
            undo_init_theta = True
            vanilla_theta = create_indep_model(self._data)
            n = vanilla_theta.shape[0]
            omega_theta = np.zeros((n + 1, n))
            omega_theta[:n] = vanilla_theta
            self._init_theta = omega_theta

        super().train(lam, maxit, trace, reltol, round_result)

        if undo_init_theta:
            self._init_theta = None
            self._result.meta["init"] = None

        return self.result

    @property
    def result(self) -> model.oMHN:
        """
        Property for retrieving the training result.

        Returns:
            model.oMHN: The resulting MHN model after training.
        """
        return self._result

    def set_device(self, device: Device):
        """
        Sets the computational device for training.

        You have three options:
            Device.AUTO: (default) automatically select the device that is likely to match the data
            Device.CPU:  use the CPU implementations to compute the scores and gradients
            Device.GPU:  use the GPU/CUDA implementations to compute the scores and gradients

        The Device enum is part of this optimizer class.

        Args:
            device (Device): The device to use (AUTO, CPU, or GPU).

        Returns:
            _Optimizer: The optimizer instance.

        Raises:
            ValueError: If the given device is not an instance of Device.
        """
        super().set_device(device)
        if device == Device.GPU:
            if cuda_available() != CUDA_AVAILABLE:
                raise CUDAError(cuda_available())

            self._gradient_and_score_func = likelihood_omhn.cuda_gradient_and_score
        else:
            self._gradient_and_score_func = {
                Device.AUTO: likelihood_omhn.gradient_and_score,
                Device.CPU: likelihood_omhn.cpu_gradient_and_score
            }[device]
        return self

    def set_penalty(
        self,
        penalty: (
            Penalty
            | tuple[
                Callable[[np.ndarray], float],
                Callable[[np.ndarray], np.ndarray],
            ]
        ),
    ):
        """
        Sets the penalty type for training.

        You have four options:
            Penalty.L1:                 (default) uses the L1 penalty as
                regularization
            Penalty.L2:                 uses the L2 penalty as
                regularization
            Penalty.SYM_SPARSE:         uses a penalty which induces
                sparsity and soft symmetry
            (penalty, penalty_deriv):   a custom penalty function and
                its derivative. They must both take the log_theta matrix
                in the form of a numpy array as input and return a float
                and a numpy array, respectively.

        The Penalty enum is part of this optimizer class.

        Args:
            penalty (Penalty | tuple[Callable[[np.ndarray], float], Callable[[np.ndarray], np.ndarray]]): The penalty to use (L1, L2, SYM_SPARSE).

        Returns:
            _Optimizer: The optimizer instance.

        Raises:
            ValueError: If the given penalty is not an instance of
            Penalty or a tuple of two functions.
        """
        if not isinstance(penalty, oMHNOptimizer.Penalty) and (
            not isinstance(penalty, tuple)
            or len(penalty) != 2
            or not callable(penalty[0])
            or not callable(penalty[1])
        ):
            raise ValueError(
                "The given penalty must either be an instance of oMHNOptimizer.Penalty or a tuple of two functions."
            )

        if isinstance(penalty, oMHNOptimizer.Penalty):
            penalty_score, penalty_gradient = {
                Penalty.L1: (penalties_omhn.l1, penalties_omhn.l1_),
                Penalty.L2: (penalties_omhn.l2, penalties_omhn.l2_),
                Penalty.SYM_SPARSE: (
                    penalties_omhn.sym_sparse,
                    penalties_omhn.sym_sparse_deriv,
                ),
            }[penalty]
        else:
            penalty_score, penalty_gradient = penalty

        self._regularized_score_func_builder = lambda grad_score_func: penalties_omhn.build_regularized_score_func(
            grad_score_func, penalty_score
        )
        self._regularized_gradient_func_builder = lambda grad_score_func: penalties_omhn.build_regularized_gradient_func(
            grad_score_func, penalty_gradient
        )

        self._penalty = (penalty_score, penalty_gradient)

        return self

class MHNType(Enum):
    """
    Enum representing the types of MHN models that can be trained.

    Attributes:
        cMHN: Classical MHN as proposed by Schill et al. (2019).
        oMHN: MHN with observation bias correction as proposed by Schill et al. (2024).
    """
    # add new types with their Optimizer classes here
    cMHN = cMHNOptimizer
    oMHN = oMHNOptimizer

    def get_optimizer(self) -> oMHNOptimizer | cMHNOptimizer:
        """Associates each enum member with its optimizer."""
        return self.value()


class Optimizer:
    """
    A dynamic wrapper for optimizer classes (e.g., oMHNOptimizer, cMHNOptimizer) that
    provides access to all methods and attributes of the wrapped optimizer instance.

    Args:
        mhn_type (MHNType, optional): Type of MHN trained by this optimizer class. Defaults to the most recent type.
    """

    # Reference to the external enums (re-export), makes separate import of enums unnecessary
    MHNType = MHNType
    Device = Device
    Penalty = Penalty

    def __new__(
        cls, mhn_type: MHNType = MHNType.oMHN
    ) -> oMHNOptimizer | cMHNOptimizer:
        if not isinstance(mhn_type, MHNType):
            mhn_type_options = ['MHNType.' + member.name for member in MHNType]
            raise ValueError(
                f"Invalid type {mhn_type}. Must be {', '.join(mhn_type_options[:-1])} or {mhn_type_options[-1]}."
            )
        return mhn_type.get_optimizer()



"""
This part of the mhn package contains functions for the posterior
sampling of MHNs with Markov Chain Monte Carlo.

kernels:
    contains the Random-Walk Metropolis, Metropolis-Adjusted Langevin
    Algorithm (MALA) and simplified manifold MALA kernels and a base
    kernel base class. Those are responsible for the actual steps of
    MCMC.

mcmc:
    contains the MCMC class that uses the kernels to create MCMC chains.
"""
# author(s): Y. Linda Hu


from . import kernels
from . import mcmc


"""
This submodule contains functions that can be used to compute the scores and gradients for the oMHN.
"""
# author(s): Stefan Vocht

from __future__ import annotations
from typing import Callable

import numpy as np

from . import likelihood_cmhn
from .state_containers import StateContainer


def _internal_gradient_and_score(omega_theta: np.ndarray, mutation_data: StateContainer, grad_and_score_func: Callable) -> tuple[np.ndarray, float]:
    """
    Computes the log-likelihood score and its gradient using the specified gradient and score function.

    Args:
        omega_theta (np.ndarray): Theta matrix for the oMHN (shape: (n+1) x n), where the last row contains observation rates.
        mutation_data (StateContainer): Data used for training.
        grad_and_score_func (Callable): Function that computes the gradient and score.

    Returns:
        tuple[np.ndarray, float]: The gradient (as a numpy array) and the score (as a float).
    """
    n = omega_theta.shape[1]
    if omega_theta.shape[0] != n+1:
        raise ValueError("omega_theta must be a 2d numpy array with n columns and n+1 rows")
    # subtract observation rates from each element in each column
    equivalent_vanilla_mhn = omega_theta[:-1] - omega_theta[-1]
    # undo changes to the diagonal
    equivalent_vanilla_mhn[range(n), range(n)] += omega_theta[-1]
    # compute the score and gradient on this theta matrix
    grad, score = grad_and_score_func(equivalent_vanilla_mhn, mutation_data)
    # compute the gradient for the observation rates
    observation_rates_gradient = -(np.sum(grad, axis=0) - grad.diagonal())
    omega_gradient = np.empty((n+1, n), dtype=np.double)
    omega_gradient[:n] = grad
    omega_gradient[-1] = observation_rates_gradient

    return omega_gradient, score


def gradient_and_score(omega_theta: np.ndarray, mutation_data: StateContainer) -> tuple[np.ndarray, float]:
    """
    Computes the log-likelihood score and gradient using both CPU and GPU.

    This function uses the CPU implementation for data points with few mutations and CUDA (if available) for data points with many mutations.
    If CUDA is not installed, the function defaults to CPU.

    Args:
        omega_theta (np.ndarray): Theta matrix for the oMHN (shape: (n+1) x n), where the last row contains observation rates.
        mutation_data (StateContainer): Data used for training.

    Returns:
        tuple[np.ndarray, float]: The gradient (as a numpy array) and the score (as a float).
    """
    return _internal_gradient_and_score(omega_theta, mutation_data, likelihood_cmhn.gradient_and_score)


def cpu_gradient_and_score(omega_theta: np.ndarray, mutation_data: StateContainer) -> tuple[np.ndarray, float]:
    """
    Computes the log-likelihood score and gradient on the CPU.

    Args:
        omega_theta (np.ndarray): Theta matrix for the oMHN (shape: (n+1) x n), where the last row contains observation rates.
        mutation_data (StateContainer): Data used for training.

    Returns:
        tuple[np.ndarray, float]: The gradient (as a numpy array) and the score (as a float).
    """
    return _internal_gradient_and_score(omega_theta, mutation_data, likelihood_cmhn.cpu_gradient_and_score)


def cpu_score(omega_theta: np.ndarray, mutation_data: StateContainer) -> float:
    """
    Computes the log-likelihood score on the CPU.

    Args:
        omega_theta (np.ndarray): Theta matrix for the oMHN (shape: (n+1) x n), where the last row contains observation rates.
        mutation_data (StateContainer): Data used for training.

    Returns:
        float: Log-likelihood score of the given oMHN for the provided data.
    """
    n = omega_theta.shape[1]
    if omega_theta.shape[0] != n+1:
        raise ValueError("omega_theta must be a 2d numpy array with n columns and n+1 rows")
    # subtract observation rates from each element in each column
    equivalent_vanilla_mhn = omega_theta[:-1] - omega_theta[-1]
    # undo changes to the diagonal
    equivalent_vanilla_mhn[range(n), range(n)] += omega_theta[-1]
    score = likelihood_cmhn.cpu_score(equivalent_vanilla_mhn, mutation_data)
    return score


def cuda_gradient_and_score(omega_theta: np.ndarray, mutation_data: StateContainer):
    """
    Computes the log-likelihood score and gradient on the GPU.

    **Note:** This function can only be used if the mhn package was compiled with CUDA.

    Args:
        omega_theta (np.ndarray): Theta matrix for the oMHN (shape: (n+1) x n), where the last row contains observation rates.
        mutation_data (StateContainer): Data used for training.

    Returns:
        tuple[np.ndarray, float]: The gradient (as a numpy array) and the score (as a float).
    """
    return _internal_gradient_and_score(omega_theta, mutation_data, likelihood_cmhn.cuda_gradient_and_score)




"""
This submodule contains the Random-Walk Metropolis, Metropolis-Adjusted
Langevin Algorithm (MALA) and simplified manifold MALA kernels and a
base kernel base class. Those are responsible for the actual steps of
MCMC.
"""
# author(s): Y. Linda Hu

from __future__ import annotations

import numpy as np
import scipy.linalg
from typing import Callable
from ..full_state_space.fisher import fisher as get_fisher
from typing import NamedTuple


class Kernel:
    """Base class for kernels used in MCMC sampling.

    Attributes:
        grad_and_log_likelihood (tuple[Callable[[np.ndarray], tuple[np.ndarray, float]]]):
            function that takes a flattened (!) log_theta matrix and
            returns a tuple of the gradient of the log-likelihood and
            the log-likelihood itself.

            **Important**: Here, the likelihood should *not* be
            normalized by the number of samples like
            in Schill et al. (2019).
        log_prior: Callable[[np.ndarray], float]: function that takes a
            flattened (!) log_theta matrix and returns the log-prior.
            This corresponds to the negative penalty term times lambda,
            multiplied by the number of samples, if applicable.
        shape: tuple[int, int]: the shape of the log_theta matrix, i.e.
            (n_events + 1, n_events) for oMHN and (n_events, n_events)
            for cMHN.
        rng: np.random.Generator | None = None: random number generator
            to be used for sampling. If None, a new RNG will be created.
    """

    def __init__(
        self,
        grad_and_log_likelihood: tuple[Callable[[np.ndarray], tuple[np.ndarray, float]]],
        log_prior: Callable[[np.ndarray], float],
        shape: tuple[int, int],
        rng: np.random.Generator | None = None,
    ):
        self.grad_and_log_likelihood = grad_and_log_likelihood
        self.log_prior = log_prior
        self.rng = rng or np.random.Generator(np.random.PCG64())
        self.shape = shape
        self.size = shape[0]*shape[1]


class smMALAResult(NamedTuple):
    """Results for one smMALAKernel step.

    Args:
        log_likelihood (float): log-likelihood of the current step
        log_prior (float): log-prior of the current step
        gradient (np.ndarray): gradient of the log-posterior at the
            current step
        G (np.ndarray): metric tensor at the current step
        cholesky (np.ndarray): Cholesky decomposition of the metric
            tensor at the current step
        mu (np.ndarray): mean of the proposal distribution at the
            current step
        det_sqrt (float): square root of the determinant of the metric
            tensor at the current step
    """
    log_likelihood: float
    log_prior: float
    gradient: np.ndarray
    G: np.ndarray
    cholesky: np.ndarray
    mu: np.ndarray
    det_sqrt: float


class smMALAKernel(Kernel):
    """Class for kernel used in simplified manifold MALA. Extends
    Kernel.

    Attributes:

        step_size (float): step size for the smMALA sampling. This is
            the :math:`\\epsilon^2/4` in the proposal distribution.
        grad_and_log_likelihood (tuple[Callable[[np.ndarray], tuple[np.ndarray, float]]]):
            function that takes a flattened (!) log_theta matrix and
            returns a tuple of the gradient of the log-likelihood and
            the log-likelihood itself.

            **Important**: Here, the likelihood should *not* be
            normalized by the number of samples like
            in Schill et al. (2019)
        log_prior: Callable[[np.ndarray], float]: function that takes a
            flattened (!) log_theta matrix and returns the log-prior.
            This corresponds to the negative penalty term times lambda,
            multiplied by the number of samples, if applicable.
        log_prior_grad: Callable[[np.ndarray], np.ndarray]: function
            that takes a flattened (!) log_theta matrix and returns the
            gradient of the log-prior.
        log_prior_hessian: Callable[[np.ndarray], np.ndarray]: function
            that takes a flattened (!) log_theta matrix and returns the
            Hessian of the log-prior.
        shape: tuple[int, int]: the shape of the log_theta matrix, i.e.
            (n_events + 1, n_events) for oMHN and (n_events, n_events)
            for cMHN.
        use_cuda: bool = False: whether to use CUDA for the computation
            of the Fisher information matrix. Only applicable if
            mhn.cuda_available() says that CUDA is available.
        rng: np.random.Generator | None = None: random number generator
            to be used for sampling. If None, a new RNG will be created.

    """

    def __init__(
        self,
        step_size: float,
        grad_and_log_likelihood: tuple[Callable[[np.ndarray], tuple[np.ndarray, float]]],
        log_prior: Callable[[np.ndarray], float],
        log_prior_grad: Callable[[np.ndarray], np.ndarray],
        log_prior_hessian: Callable[[np.ndarray], np.ndarray],
        shape: tuple[int, int],
        use_cuda: bool = False,
        rng: np.random.Generator | None = None,
    ):

        super().__init__(grad_and_log_likelihood=grad_and_log_likelihood,
                         log_prior=log_prior, shape=shape, rng=rng)

        self.step_size = step_size
        self.log_prior_grad = log_prior_grad
        self.log_prior_hessian = log_prior_hessian
        self.use_cuda = use_cuda

    def propose(self, prev_step: np.ndarray, prev_step_res: smMALAResult
                ) -> tuple[np.ndarray, smMALAResult]:
        """Propose a new step based on the previous step and its
        results.
        This is done according to

        :math:`\\mathcal N\\bigg(\\theta + \\frac{\\epsilon}{2} G(\\theta)^{-1} \\frac{\\partial \\log p(\\theta)}{\\partial \\theta}, \\epsilon G(\\theta)^{-1}\\bigg)`
        
        where
        
        :math:`G(\\theta) = I(\\theta) - H(\\log \\pi(\\theta))`
        
        with :math:`I` the Fisher information matrix and :math:`H` the Hessian of the log-prior.


        Args:
            prev_step (np.array): the previous step
            prev_step_res (smMALAResult): Results of the previous step.

        Returns:
            tuple[np.ndarray, smMALAResult]: the new step and its
            results.
        """

        # draw random normal number
        z = self.rng.normal(size=prev_step.size)

        # transform with inverse transformed cholesky matrix
        y = np.sqrt(self.step_size) * scipy.linalg.solve_triangular(
            prev_step_res.cholesky.T, z, lower=False
        )

        new_step = prev_step_res.mu + y
        return new_step, self.get_params(new_step)

    def log_accept(
            self,
            prev_step: np.ndarray,
            prev_step_res: smMALAResult,
            new_step: np.ndarray,
            new_step_res: smMALAResult) -> float:
        """Give acceptance ratio of accepting the new step.

        This is given by

        :math:`\\frac{p(\\theta' | D) q(\\theta | \\theta')\\pi(\\theta')}{p(\\theta|D) q(\\theta' | \\theta) \\pi(\\theta)}`
        
        where :math:`p(.| D)` is the likelihood, :math:`\\pi(.)` is the
        prior and :math:`q(.|.)` is the proposal distribution.

        Args:
            prev_step (np.ndarray): the previous step
            prev_step_res (smMALAResult): Results of the previous step.
            new_step (np.ndarray): the new step
            new_step_res (smMALAResult): Results of the new step.

        Returns:
            float: the acceptance ratio
        """

        # p(theta' | D) q(theta | theta') pr(theta') / p(theta|D) q(theta' | theta) pr(theta)

        acceptance_ratio = (
            new_step_res.log_likelihood
            + new_step_res.log_prior
            + self.log_q(
                theta_proposed=prev_step,
                G=new_step_res.G,
                det_sqrt_G=new_step_res.det_sqrt,
                mu=new_step_res.mu,
            )
            - prev_step_res.log_likelihood
            - prev_step_res.log_prior
            - self.log_q(
                theta_proposed=new_step,
                G=prev_step_res.G,
                det_sqrt_G=prev_step_res.det_sqrt,
                mu=prev_step_res.mu,
            )
        )

        return acceptance_ratio

    def one_step(
        self,
        prev_step: np.ndarray,
        prev_step_res: smMALAResult,
        return_info: bool = False
    ) -> tuple[np.ndarray, smMALAResult] | tuple[np.ndarray, smMALAResult, float, int]:
        """Perform one smMALA step.

        Args:
            prev_step (np.ndarray): the previous step
            prev_step_res (smMALAResult): Results of the previous
                step.
            return_info (bool, optional): Whether to return the
                acceptance ratio and acceptance indicator. Defaults to
                False.

        Returns:
            tuple[np.ndarray, smMALAResult] |
            tuple[np.ndarray, smMALAResult, float, int]: The new step
                and its results, and optionally acceptance ratio and
                acceptance indicator.
        """

        new_step, new_step_res = self.propose(prev_step, prev_step_res)
        acceptance_ratio = self.log_accept(
            prev_step, prev_step_res, new_step, new_step_res
        )

        if np.log(self.rng.random()) < acceptance_ratio:
            if return_info:
                return (new_step, new_step_res, acceptance_ratio, 1)
            return (new_step, new_step_res)
        else:
            if return_info:
                return (prev_step, prev_step_res, acceptance_ratio, 0)
            return (prev_step, prev_step_res)

    def get_params(self, initial_step: np.ndarray) -> smMALAResult:
        """Get the results for a given step.

        Args:
            initial_step (np.ndarray): The step for which to get the
                results.

        Returns:
            smMALAResult: The results for the given step.
        """

        # Get gradient, likelihood and G matrix for new theta
        log_likelihood_grad, log_likelihood = self.grad_and_log_likelihood(
            initial_step)
        log_prior = self.log_prior(initial_step)

        log_posterior_grad = log_likelihood_grad + \
            self.log_prior_grad(initial_step)

        fisher = get_fisher(
            log_theta=initial_step.reshape(self.shape),
            omhn=self.shape[0] == self.shape[1] + 1,
            use_cuda=self.use_cuda,
        )
        G = fisher - self.log_prior_hessian(initial_step)
        try:
            cholesky = scipy.linalg.cholesky(G, lower=True)
        except np.linalg.LinAlgError as e:
            raise np.linalg.LinAlgError(
                "The cholesky decomposition of the metric tensor" \
                "failed. This likely means that the metric tensor is " \
                "not positive definite. Check that the hessian of " \
                "the log-prior is negative definite and that the " \
                "and, if necessary, decrease the stepsize."
            ) from e
        det_sqrt = np.diag(cholesky).prod()

        # Get mu, the mean of the proposal distribution w.r.t. the new theta
        # this is log_theta + 0.5 * STEP_SIZE * G^-1 * gradient
        y = scipy.linalg.solve_triangular(
            cholesky.T,
            scipy.linalg.solve_triangular(
                cholesky,
                log_posterior_grad.flatten(),
                lower=True,
            ),
            lower=False,
        )
        mu = initial_step.flatten() + 0.5 * self.step_size * y

        return smMALAResult(
            log_likelihood=log_likelihood,
            log_prior=log_prior,
            gradient=log_posterior_grad,
            G=G,
            cholesky=cholesky,
            mu=mu,
            det_sqrt=det_sqrt,
        )

    def log_q(
        self,
        theta_proposed: np.ndarray,
        G: np.ndarray,
        det_sqrt_G: float,
        mu: np.ndarray,
    ) -> float:
        """Compute the logarithm of the proposal distribution density
            q(theta_new | theta) for the smMALA algorithm.
            This is according to https://en.wikipedia.org/wiki/Multivariate_normal_distribution
            #Density_function.

        Args:
            theta_proposed (np.ndarray): New proposed theta
            G (np.ndarray): Metric tensor w.r.t. the old theta
            det_sqrt_G (float): Square root of the determinant of the
                metric tensor w.r.t. the old theta
            mu (np.ndarray): Mean of the proposal distribution w.r.t.
                the old theta

        Returns:
            float: The logarithm of the proposal distribution
                q(theta_new | theta) density
        """
        # we can leave out the constant factor
        # (2 * np.pi) ** (n_events**2 )
        # in the denominator, as well as the scaling
        # STEP_SIZE ** (n_events**2)
        # 1/sqrt(det(G^-1)) = sqrt(det(G))
        return -0.5 * (theta_proposed - mu).T @ G @ (
            theta_proposed - mu
        ) / self.step_size + np.log(det_sqrt_G)


class MALAResult(NamedTuple):
    """Results for one smMALAKernel step.

    Args:
        log_likelihood (float): log-likelihood of the current step
        log_prior (float): log-prior of the current step
        mu (np.ndarray): mean of the proposal distribution at the
            current step
    """
    log_likelihood: float
    log_prior: float
    mu: np.ndarray


class MALAKernel(Kernel):
    """Class for kernel used in the Metropolis-Adjusted Langevin
    Algorithm. Extends Kernel.

    Attributes:
        step_size (float): step size for the MALA sampling.
        grad_and_log_likelihood (tuple[Callable[[np.ndarray], tuple[np.ndarray, float]]]):
            function that takes a flattened (!) log_theta matrix and
            returns a tuple of the gradient of the log-likelihood and
            the log-likelihood itself.

            **Important**: Here, the likelihood should *not* be
            normalized by the number of samples like
            in Schill et al. (2019)
        log_prior: Callable[[np.ndarray], float]: function that takes a
            flattened (!) log_theta matrix and returns the log-prior.
            This corresponds to the negative penalty term times lambda,
            multiplied by the number of samples, if applicable.
        log_prior_grad: Callable[[np.ndarray], np.ndarray]: function
            that takes a flattened (!) log_theta matrix and returns the
            gradient of the log-prior.
        shape: tuple[int, int]: the shape of the log_theta matrix, i.e.
            (n_events + 1, n_events) for oMHN and (n_events, n_events)
            for cMHN.
        rng: np.random.Generator | None = None: random number generator
            to be used for sampling. If None, a new RNG will be created.

    """

    def __init__(
        self,
        step_size: float,
        grad_and_log_likelihood: tuple[Callable[[np.ndarray], tuple[np.ndarray, float]]],
        log_prior: Callable[[np.ndarray], float],
        log_prior_grad: Callable[[np.ndarray], np.ndarray],
        shape: tuple[int, int],
        rng: np.random.Generator | None = None,
    ):

        super().__init__(grad_and_log_likelihood=grad_and_log_likelihood,
                         log_prior=log_prior, shape=shape, rng=rng)

        self.step_size = step_size
        self.log_prior_grad = log_prior_grad

    def propose(self, prev_step: np.ndarray, prev_step_res: MALAResult
                ) -> tuple[np.ndarray, MALAResult]:
        """Propose a new step based on the previous step and its 
        results.

        Args:
            prev_step (np.array): the previous step
            prev_step_res (MALAResult): Results of the previous step.

        Returns:
            tuple[np.ndarray, MALAResult]: the new step and its
            results.
        """
        z = self.rng.normal(size=self.size)
        new_step = prev_step_res.mu + np.sqrt(self.step_size) * z

        return new_step, self.get_params(new_step)

    def log_accept(
            self,
            prev_step: np.ndarray,
            prev_step_res: MALAResult,
            new_step: np.ndarray,
            new_step_res: MALAResult) -> float:
        """Give acceptance ratio of accepting the new step.

        Args:
            prev_step (np.ndarray): the previous step
            prev_step_res (MALAResult): Results of the previous step.
            new_step (np.ndarray): the new step
            new_step_res (MALAResult): Results of the new step.

        Returns:
            float: the acceptance ratio
        """
        # p(theta' | D) q(theta | theta') pr(theta') /
        # p(theta|D) q(theta' | theta) pr(theta)

        acceptance_ratio = (
            new_step_res.log_likelihood
            + new_step_res.log_prior
            + self.log_q(
                theta_proposed=prev_step,
                mu=new_step_res.mu,
            )
            - prev_step_res.log_likelihood
            - prev_step_res.log_prior
            - self.log_q(
                theta_proposed=new_step,
                mu=prev_step_res.mu,
            )
        )

        return acceptance_ratio

    def one_step(
        self,
        prev_step: np.ndarray,
        prev_step_res: MALAResult,
        return_info: bool = False
    ) -> tuple[np.ndarray, MALAResult] | tuple[np.ndarray, MALAResult, float, int]:
        """Perform one MALA step.

        Args:
            prev_step (np.ndarray): the previous step
            prev_step_res (MALAResult): Results of the previous step.
            return_info (bool, optional): Whether to return the
                acceptance ratio and acceptance indicator. Defaults to
                False.

        Returns:
            tuple[np.ndarray, MALAResult] |
            tuple[np.ndarray, MALAResult, float, int]: The new step and
                its results, and optionally acceptance ratio and
                acceptance indicator.
        """
        new_step, new_step_res = self.propose(prev_step, prev_step_res)
        acceptance_ratio = self.log_accept(
            prev_step, prev_step_res, new_step, new_step_res
        )

        if np.log(self.rng.random()) < acceptance_ratio:
            if return_info:
                return (new_step, new_step_res, acceptance_ratio, 1)
            return (new_step, new_step_res)
        else:
            if return_info:
                return (prev_step, prev_step_res, acceptance_ratio, 0)
            return (prev_step, prev_step_res)

    def get_params(self, initial_step: np.ndarray) -> MALAResult:
        """Get the results for a given step.

        Args:
            initial_step (np.ndarray): The step for which to get the
                results.

        Returns:
            MALAResult: The results for the given step.
        """

        log_likelihood_grad, log_likelihood = self.grad_and_log_likelihood(
            initial_step)
        log_posterior_grad = log_likelihood_grad + \
            self.log_prior_grad(initial_step.reshape(self.shape))

        mu = initial_step.flatten() \
            + 0.5 * self.step_size * log_posterior_grad.flatten()
        return MALAResult(
            log_likelihood=log_likelihood,
            log_prior=self.log_prior(initial_step.reshape(self.shape)),
            mu=mu,
        )

    def log_q(
        self,
        theta_proposed: np.ndarray,
        mu: np.ndarray,
    ) -> float:
        """Compute the logarithm of the proposal distribution density
        q(theta_new | theta) for the MMALA algorithm.
        This is according to https://en.wikipedia.org/wiki/Multivariate_normal_distribution#Density_function.

        Args:
            theta_new (np.ndarray): New proposed theta
            G (np.ndarray): Metric tensor w.r.t. the old theta
            det_sqrt_G (float): Square root of the determinant of the
            metric tensor w.r.t. the old theta
            mu (np.ndarray): Mean of the proposal distribution w.r.t.
            the old theta

        Returns:
            float: The logarithm of the proposal distribution 
            q(theta_new | theta) density
        """
        # we can leave out the constant factor (2 * np.pi) ** (n_events**2 )
        # in the denominator, as well as the scaling STEP_SIZE ** (n_events**2)
        return -0.5 * np.sum((theta_proposed - mu) ** 2) / self.step_size


class RWMResult(NamedTuple):
    """Results for one RWMKernel step.

    Args:
        log_likelihood (float): log-likelihood of the current step
        log_prior (float): log-prior of the current step
    """
    log_likelihood: float
    log_prior: float


class RWMKernel(Kernel):
    """Class for kernel used in simplified Random-Walk Metropolis.
    Extends Kernel.

    Attributes:
        grad_and_log_likelihood (tuple[Callable[[np.ndarray], tuple[np.ndarray, float]]]):
            function that takes a flattened (!) log_theta matrix and
            returns a tuple of the gradient of the log-likelihood and
            the log-likelihood itself.

            **Important**: Here, the likelihood should *not* be
            normalized by the number of samples like
            in Schill et al. (2019)
        log_prior: Callable[[np.ndarray], float]: function that takes a
            flattened (!) log_theta matrix and returns the log-prior.
            This corresponds to the negative penalty term times lambda,
            multiplied by the number of samples, if applicable.
        scale (float): scale of the normal proposal distribution.
        shape: tuple[int, int]: the shape of the log_theta matrix, i.e.
            (n_events + 1, n_events) for oMHN and (n_events, n_events)
            for cMHN.
        rng: np.random.Generator | None = None: random number generator
            to be used for sampling. If None, a new RNG will be created.

    """

    def __init__(
        self,
        grad_and_log_likelihood: tuple[Callable[[np.ndarray], np.ndarray], Callable[[np.ndarray], float]],
        log_prior: Callable[[np.ndarray], float],
        step_size: float,
        shape: tuple[int, int],
        rng: np.random.Generator | None = None,
    ):
        super().__init__(grad_and_log_likelihood=grad_and_log_likelihood,
                         log_prior=log_prior, shape=shape, rng=rng)

        self.step_size = step_size
        self.sigma = step_size * np.eye(self.size)

    def propose(self, prev_step: np.ndarray, prev_step_res: RWMResult):
        """Propose a new step based on the previous step and its
        results.
        The proposal is made according to a normal distribution centered
        at the previous step with standard deviation `step_size`.

        Args:
            prev_step (np.array): the previous step
            prev_step_res (RWMResult): Results of the previous step.

        Returns:
            tuple[np.ndarray, RWMResult]: the new step and its
            results.
        """
        # draw random normal number
        new_step = self.rng.normal(
            loc=prev_step,
            scale=self.step_size,
            size=self.size,
        )

        return new_step, self.get_params(new_step)

    def log_accept(
            self,
            prev_step: np.ndarray,
            prev_step_res: RWMResult,
            new_step: np.ndarray,
            new_step_res: RWMResult) -> float:
        """Give acceptance ratio of accepting the new step. This is
        given by
        p(new_step | D) pr(new_step) / p(prev_step|D) pr(prev_step)
        where p(.| D) is the likelihood and pr(.) is the prior.

        Args:
            prev_step (np.ndarray): the previous step
            prev_step_res (RWMResult): Results of the previous step.
            new_step (np.ndarray): the new step
            new_step_res (RWMResult): Results of the new step.

        Returns:
            float: the acceptance ratio
        """
        # p(theta' | D) pr(theta') / p(theta|D) pr(theta)

        acceptance_ratio = (
            new_step_res.log_likelihood
            + new_step_res.log_prior
            - prev_step_res.log_likelihood
            - prev_step_res.log_prior
        )

        return acceptance_ratio

    def one_step(
            self,
            prev_step: np.ndarray,
            prev_step_res: RWMResult,
            return_info: bool = False):
        """Perform one RWM step.

        Args:
            prev_step (np.ndarray): the previous step
            prev_step_res (RWMResult): Results of the previous
                step.
            return_info (bool, optional): Whether to return the
                acceptance ratio and acceptance indicator. Defaults to
                False.

        Returns:
            tuple[np.ndarray, RWMResult] |
            tuple[np.ndarray, RWMResult, float, int]: The new step
            and its results, and optionally acceptance ratio and
            acceptance indicator.
        """

        new_step, new_step_res = self.propose(prev_step, prev_step_res)
        acceptance_ratio = self.log_accept(
            prev_step, prev_step_res, new_step, new_step_res
        )

        if np.log(self.rng.random()) < acceptance_ratio:
            if return_info:
                return (new_step, new_step_res, acceptance_ratio, 1)
            return (new_step, new_step_res)
        else:
            if return_info:
                return (prev_step, prev_step_res, acceptance_ratio, 0)
            return (prev_step, prev_step_res)

    def get_params(self, initial_step):
        """Get the results for a given step.

        Args:
            initial_step (np.ndarray): The step for which to get the
                results.

        Returns:
            RWMResult: The results for the given step.
        """
        return RWMResult(
            self.grad_and_log_likelihood(initial_step.reshape(self.shape))[1],
            self.log_prior(initial_step.reshape(self.shape))
        )



import unittest
import warnings
import numpy as np
import mhn.full_state_space.fisher
from mhn.full_state_space import Likelihood, ModelConstruction
from mhn.training.likelihood_cmhn import gradient_and_score
from mhn.training.state_containers import StateContainer


class TestFisher(unittest.TestCase):

    def test_cython_fisher(self):
        """Test the Fisher information matrix for cMHN.
        """
        n = 4
        log_theta = ModelConstruction.random_theta(n)

        p_th = Likelihood.generate_pTh(log_theta)
        FIM = np.zeros((n * n, n * n))
        X = [[l, k, j, i] for i in range(2) for j in range(2)
             for k in range(2) for l in range(2)]

        for i, x in enumerate(X):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                grad = gradient_and_score(
                    log_theta,
                    StateContainer(
                        np.array(x, dtype=np.int32).reshape(1, n)
                    ))[0].flatten()
            FIM += p_th[i] * np.outer(grad, grad)

        for use_cuda in [True, False]:
            with self.subTest(use_cuda):
                if use_cuda:
                    if mhn.cuda_available() != \
                            mhn.CUDA_AVAILABLE:
                        self.skipTest("CUDA not available for testing")
                np.testing.assert_allclose(
                    FIM,
                    mhn.full_state_space.fisher.fisher(
                        log_theta=log_theta, omhn=False, use_cuda=use_cuda)
                )

    def test_omhn_fisher_theta_theta(self):
        """Test upper left Block of oMHN Fisher information matrix."""

        n = 6
        o_log_theta = 2 * np.random.random((n + 1, n)) - 1
        c_log_theta = mhn.model.oMHN(
            o_log_theta).get_equivalent_classical_mhn().log_theta

        np.testing.assert_allclose(
            mhn.full_state_space.fisher.cython_fisher(c_log_theta),
            mhn.full_state_space.fisher.omhn_fisher(o_log_theta)[:n**2, :n**2]
        )

    def test_omhn_fisher_omega_theta(self):
        """Test off-diagonal Blocks of oMHN Fisher information matrix."""

        n = 6

        log_theta = 2 * np.random.random((n + 1, n)) - 1
        fisher_matrix = mhn.full_state_space.fisher.omhn_fisher(log_theta)

        np.testing.assert_allclose(
            fisher_matrix[n**2:, :n**2],
            fisher_matrix[:n**2, n**2:].T)

        fisher_omega_theta = fisher_matrix[n**2:, :n**2]

        for j in range(n):
            for s in range(n):
                for t in range(n):
                    np.testing.assert_allclose(
                        fisher_omega_theta[j, s * n + t],
                        -sum(fisher_matrix[i * n + j, s * n + t]
                             for i in range(n) if i != j)
                    )

    def test_omhn_fisher_omega_omega(self):
        """Test lower right Block of oMHN Fisher information matrix."""

        n = 6
        log_theta = 2 * np.random.random((n + 1, n)) - 1
        fisher_matrix = mhn.full_state_space.fisher.omhn_fisher(log_theta)

        fisher_omega_omega = fisher_matrix[n**2:, n**2:]

        np.testing.assert_allclose(
            fisher_omega_omega,
            fisher_omega_omega.T
        )

        for j in range(n):
            for t in range(n):
                np.testing.assert_allclose(
                    fisher_omega_omega[j, t],
                    sum(
                        sum(fisher_matrix[i * n + j, s * n + t]
                            for s in range(n) if s != t)
                        for i in range(n) if i != j
                    ))


if __name__ == "__main__":
    unittest.main()



"""
This submodule implements UtilityFunctions.R from the original R implementation in Python.

It contains functions useful for preprocessing training data.
"""
# author(s): Stefan Vocht

import numpy as np


def state_to_int(x: np.ndarray) -> int:
    """
    Interprets a binary array as a binary number and returns the corresponding integer value.

    Args:
        x (np.ndarray): Binary array, typically representing a row of the mutation matrix in this context.

    Returns:
        int: The integer value corresponding to the binary array.
    """

    # reverse list and convert elements to string
    x = map(str, x[::-1])
    return int(''.join(x), 2)


def data_to_pD(data: np.ndarray) -> np.ndarray:
    """
    Calculates the probability distribution for the different events from a given binary mutation matrix.

    Args:
        data (np.ndarray): A numpy array or matrix representing the mutation data.

    Returns:
        np.ndarray: The probability distribution of the different events.
    """

    n = data.shape[1]
    N = 2**n

    # convert data into a list of integers, where each number represents a different event
    data = list(map(state_to_int, data))

    # calculate the probability distribution
    pD = np.bincount(data, minlength=N)
    pD = pD / np.sum(pD)

    return pD


def finite_sample(p_th: np.ndarray, k: int) -> np.ndarray:
    """
    Generates a random sample given a probability distribution and returns the probability distribution for the new sample.

    Args:
        p_th (np.ndarray): Probability distribution of events (the distribution of a true Theta).
        k (int): The number of samples to generate.

    Returns:
        np.ndarray: The probability distribution of events from the generated samples.
    """
    n = p_th.size

    return np.bincount(np.random.choice(n, k, replace=True, p=p_th), minlength=n) / k


def KL_div(p: np.ndarray, q: np.ndarray) -> float:
    """
    Computes the Kullback–Leibler divergence between two probability distributions.

    Args:
        p (np.ndarray): Probability distribution p.
        q (np.ndarray): Probability distribution q.

    Returns:
        float: The KL-divergence of p and q.
    """
    return p.dot(np.log(p)) - p.dot(np.log(q))



"""
This submodule contains the MCMC class that runs Markov Chain Monte
Carlo sampling for MHNs.

"""

# author: Y. Linda Hu

from __future__ import annotations

from ..optimizers import Optimizer, oMHNOptimizer, cMHNOptimizer, Penalty
from ..model import oMHN, cMHN
from ..training.likelihood_cmhn \
    import gradient_and_score as cmhn_grad_and_log_likelihood
from ..training.likelihood_omhn \
    import gradient_and_score as omhn_grad_and_log_likelihood
from ..training.state_containers import StateContainer
from typing import Callable, Literal
import numpy as np
import multiprocessing as mp
from .kernels import Kernel, smMALAKernel, RWMKernel, MALAKernel
from ..training import penalties_cmhn, penalties_omhn
import warnings
import arviz

class MCMC:
    """Markov chain Monte Carlo sampler for MHN.

    The simplest way to create an MCMC sampler is to provide a trained
    ``Optimizer``. This already includes the dataset, the regularization
    strength lambda, the penalty and its gradient.

    This is enough to create an RWM or MALA sampler. For an smMALA
    sampler, the Hessian of the penalty is also needed, which is not
    stored in the optimizer. In this case, please provide the penalty
    and its derivatives directly via ``penalty`` or ``log_prior``.

    Alternatively, you can create an MCMC sampler from a trained MHN
    model, the data used to train it, and the penalty used for training.
    Depending on the Kernel (RWM, MALA, or smMALA), you have to provide
    the penalty, its gradient, and its Hessian,

    - by providing ``penalty`` either as a :class:`Penalty`
      (only for RWM/MALA) or as a tuple of callables of length 1, 2,
      or 3, or
    - by providing ``log_prior`` as a tuple of callables of length 1,
      2, or 3.
    The difference between ``penalty`` and ``log_prior`` is that the
    former is unscaled by :math:`\\lambda` and positive (such as it is
    stored in an ``Optimizer`` object), while the latter is scaled by
    lambda and negative.

    When using a custom penalty or prior, initial chain values must be
    set manually by setting ``Sampler.initial_step`` to an array of
    shape (n_chains, 1, m) with m the number of parameters.

    Args:
        optimizer (Optimizer, optional): Trained Optimizer.
        mhn_model (oMHN | cMHN, optional): MHN model.
        data (np.ndarray, optional): Data used to train the MHN model.
        penalty (Penalty | tuple[Callable[[np.ndarray], float], Callable[[np.ndarray], np.ndarray]], optional):
            Penalty used during training. If not Penalty, ``penalty[0]``
            gives the penalty (positive and unscaled by
            :math:`\\lambda`), ``penalty[1]`` its gradient and
            ``penalty[2]`` its Hessian.
        log_prior: Log-prior used during training. ``log_prior[0]``
            gives the log_prior (negative and scaled by 
            :math:`\\lambda`), ``penalty[1]`` its gradient and
            ``penalty[2]`` its Hessian.
        n_chains (int, optional): Number of parallel chains to run.
            Defaults to ``10``.
        step_size: (Literal["auto"] | int | np.ndarray | ) Stepsize of the Kernel. See the documentation of the
            kernels (:class:`RWMKernel<mhn.mcmc.kernels.RWMKernel>`,
            :class:`MALAKernel<mhn.mcmc.kernels.MALAKernel>`,
            :class:`smMALAKernel<mhn.mcmc.kernels.smMALAKernel>`) for
            more details. If ``"auto"``, the stepsize is tuned
            automatically before the first run using
            :func:`tune_stepsize`. If array, every chain uses a
            different step size. Defaults to ``"auto"``.
        kernel_class (Kernel, optional): Kernel class to use for MCMC
            sampling. Must be one of ``RWMKernel``, ``MALAKernel``,
            ```smMALAKernel``. Defaults to ``MALAKernel``.
        thin (int, optional): Thinning factor for MCMC sampling.
            Defaults to ``100``.
        seed (int, optional): Random seed for reproducibility. Defaults
            to ``None``.
    """

    _kernel_args = {
        MALAKernel: ["log_prior_grad"],
        smMALAKernel: ["log_prior_grad", "log_prior_hessian"],
        RWMKernel: []
    }

    _penalty_dict = {
        "cmhn": {
            Penalty.L1: (penalties_cmhn.l1, penalties_cmhn.l1_),
            Penalty.L2: (penalties_cmhn.l2, penalties_cmhn.l2_),
            Penalty.SYM_SPARSE: (
                penalties_cmhn.sym_sparse,
                penalties_cmhn.sym_sparse_deriv,
            ),
        },
        "omhn": {
            Penalty.L1: (penalties_omhn.l1, penalties_omhn.l1_),
            Penalty.L2: (penalties_omhn.l2, penalties_omhn.l2_),
            Penalty.SYM_SPARSE: (
                penalties_omhn.sym_sparse,
                penalties_omhn.sym_sparse_deriv,
            ),
        },
    }

    def __init__(
            self, *,
            optimizer: Optimizer = None,
            mhn_model: oMHN | cMHN | None = None,
            data: np.ndarray | None = None,
            penalty: Penalty | tuple[Callable] | None = None,
            log_prior: tuple[Callable] | None = None,
            n_chains: int = 10,
            step_size: Literal["auto"] | float | np.ndarray = "auto",
            kernel_class: Kernel = MALAKernel,
            thin: int = 100,
            seed: int | None = None,) -> None:

        # Model initialization

        if optimizer is None:

            self.optimizer = None

            if mhn_model is None or data is None or\
                    (penalty is None and log_prior is None):
                raise ValueError(
                    "When optimizer is not provided, mhn_model, data," +
                    "and penalty or log_prior must be provided."
                )

            self._bin_datamatrix = data
            self.data = StateContainer(data) if isinstance(
                data, np.ndarray) else data
            self.n_samples = self.data.get_data_shape()[0]
            self.shape = mhn_model.log_theta.shape
            self.size = mhn_model.log_theta.size

            if isinstance(mhn_model, oMHN):
                self.omhn = True
            elif isinstance(mhn_model, cMHN):
                self.omhn = False
            else:
                raise ValueError("mhn_model must be an oMHN or cMHN "
                                 "instance.")
            assert hasattr(mhn_model, "meta"), "mhn_model must be " \
                "supplied with metadata."
            try:
                self.lam = mhn_model.meta["lambda"] * self.n_samples
            except AttributeError:
                raise ValueError("mhn_model metadata must include "
                                 "lambda.")

        else:

            if isinstance(optimizer, oMHNOptimizer):
                self.omhn = True
            elif isinstance(optimizer, cMHNOptimizer):
                self.omhn = False
            else:
                raise ValueError("optimizer must be an oMHNOptimizer "
                                 "or cMHNOptimizer instance.")
            self.optimizer = optimizer
            if mhn_model is not None or data is not None:
                raise ValueError(
                    "When optimizer is provided, mhn_model and data "
                    "must not be provided."
                )

            try:
                self.lam = optimizer.result.meta["lambda"] * \
                    optimizer._data.get_data_shape()[0]
                self.data = optimizer._data
                self._bin_datamatrix = optimizer._bin_datamatrix
                self.shape = optimizer.result.log_theta.shape
                self.size = optimizer.result.log_theta.size
            except AttributeError:
                raise ValueError(
                    "optimizer must have loaded data and must have "
                    "been trained.")

            self.n_samples = self.data.get_data_shape()[0]

            if penalty is None and log_prior is None:

                if not hasattr(optimizer, "_penalty")\
                        or optimizer._penalty is None:
                    raise ValueError(
                        "optimizer does not have a penalty, provide "
                        "either penalty or log_prior")

                else:

                    if kernel_class == smMALAKernel:
                        raise ValueError(
                            "When using smMALA kernel, the penalty "
                            "included in the optimizer is not "
                            "sufficient, because no Hessian is "
                            "stored. Provide penalty or log_prior "
                            "including its 1st and 2nd derivative.")
                    else:
                        penalty = optimizer._penalty

            else:

                if hasattr(optimizer, "_penalty")\
                        and optimizer._penalty is not None:

                    warnings.warn(
                        "The optimizers penalty is overwritten by the "
                        "provided penalty or log_prior.", UserWarning
                    )

        if penalty is not None and log_prior is not None:
            raise ValueError("Provide only one of penalty or log_prior.")

        self._log_prior = log_prior

        if penalty is not None:

            if isinstance(penalty, Penalty):
                penalty = self._penalty_dict[
                    "omhn" if self.omhn else "cmhn"][penalty]

            log_prior = (None, None, None)
            
        self._penalty = penalty

        log_prior = log_prior or (None, None, None)

        self.log_prior_grad, self.log_prior_hessian = None, None
        self.log_prior = log_prior[0] or self._get_log_prior(
            penalty[0])
        if kernel_class in [MALAKernel, smMALAKernel]:
            self.log_prior_grad = log_prior[1] or self._get_log_prior_grad(
                penalty[1])
            if kernel_class == smMALAKernel:
                self.log_prior_hessian = log_prior[2] or \
                    self._get_log_prior_hessian(penalty[2])

        # Sampler initialization

        self.n_chains = n_chains
        self.log_thetas = np.array([]).reshape(n_chains, 0, self.size)
        self.step_size = step_size
        self.thin = thin
        self.initial_step = None

        seed_sequence = np.random.SeedSequence(seed)
        self.rng = np.random.Generator(
            np.random.PCG64(seed_sequence.spawn(1)[0]))
        self.kernel_rngs = [
            np.random.Generator(
                np.random.PCG64(sese),
            )
            for sese in seed_sequence.spawn(self.n_chains)
        ]
        self.grad_and_log_likelihood = self._get_grad_and_log_likelihood()
        self.kernel_class = kernel_class

    def _get_grad_and_log_likelihood(
            self) -> Callable[[np.ndarray], tuple[np.ndarray, float]]:
        """
        Get the grad_and_log_likelihood function for an MHN. The default
        gradient and likelihood in the mhn.training module are
        normalized by the dataset size. This is reversed here
        """

        if self.omhn:

            def grad_and_log_likelihood(log_theta: np.ndarray) \
                    -> tuple[np.ndarray, float]:
                grad, log_likelihood = omhn_grad_and_log_likelihood(
                    omega_theta=log_theta.reshape(self.shape),
                    mutation_data=self.data,
                )
                return self.n_samples * grad, self.n_samples * log_likelihood

        else:

            def grad_and_log_likelihood(log_theta: np.ndarray) \
                    -> tuple[np.ndarray, float]:
                grad, log_likelihood = cmhn_grad_and_log_likelihood(
                    log_theta=log_theta.flatten(),
                    data_matrix=self.data,
                )
                return self.n_samples * grad, self.n_samples * log_likelihood

        return grad_and_log_likelihood

    def _get_log_prior(
        self, penalty: Callable[[np.ndarray], float]
    ) -> Callable[[np.ndarray], float]:
        """Get the log_prior as -lam * penalty, where lam is the
        regularization strength from MHN training.

        Args:
            penalty (Callable[[np.ndarray], float]): The penalty
            function used for MHN training.
        """

        def log_prior(log_theta: np.ndarray) -> float:
            return -self.lam * penalty(log_theta)

        return log_prior

    def _get_log_prior_grad(
        self, penalty_grad: Callable[[np.ndarray], np.ndarray]
    ) -> Callable[[np.ndarray], np.ndarray]:
        """Get the log_prior_grad as -lam * penalty_grad, where lam is
        the regularization strength from MHN training.

        Args:
            penalty_grad (Callable[[np.ndarray], np.ndarray]): The
                gradient of the penalty function used for MHN training.
        """

        def log_prior_grad(log_theta: np.ndarray) -> float:
            return -self.lam * penalty_grad(log_theta)

        return log_prior_grad

    def _get_log_prior_hessian(
        self, penalty_hessian: Callable[[np.ndarray], np.ndarray]
    ) -> Callable[[np.ndarray], np.ndarray]:
        """Get the log_prior_hessian as -lam * penalty_hessian, where
        lam is the regularization strength from MHN training.

        Args:
            penalty_hessian (Callable[[np.ndarray], np.ndarray]): The
                hessian of the penalty function used for MHN training.
        """

        def log_prior_hessian(log_theta: np.ndarray) -> float:
            return -self.lam * penalty_hessian(log_theta)

        return log_prior_hessian

    def _take_initial_step(self):
        """Take the initial step for MCMC sampling. Only works if
        penalty is one of the predefined penalties for
        ``mhn.training.optimizer``.

        Raises:
            NotImplementedError: When using a custom penalty, initial
                chain values must be set manually by setting
                ``Sampler.initial_step`` to an array of shape
                (n_chains, 1, m) with m the number of parameters.
        """
        if self.initial_step is not None:
            return

        if self._penalty in [Penalty.L2, self._penalty_dict["omhn" if self.omhn else "cmhn"][Penalty.L2]]:
            self.initial_step = self.rng.normal(
                size=(self.n_chains, 1, self.size),
                scale=1 / np.sqrt(2 * self.lam),
            )

        elif self._penalty in [
            Penalty.L1,
            Penalty.SYM_SPARSE,
            self._penalty_dict["omhn" if self.omhn else "cmhn"][Penalty.L1],
            self._penalty_dict["omhn" if self.omhn else "cmhn"][Penalty.SYM_SPARSE]
        ]:
            self.initial_step = self.rng.laplace(
                size=(self.n_chains, 1, self.size),
                scale=1 / self.lam,
            )

        else:
            raise NotImplementedError(
                "When using a custom penalty, you must manually set " +
                "the initial chain values by setting " +
                R"`Sampler.initial_step` to an array of shape " +
                "(n_chains, 1, m) with m the number of parameters. "
            )

    def _walker(
        self,
        prev_step: np.ndarray,
        walker_id: int,
        n_steps: int,
        verbose: bool,
    ) -> tuple[int, np.ndarray, np.random.Generator]:
        """Internal walker function. Performs the MCMC sampling for one
        chain.

        Args:
            prev_step (np.ndarray): Initial value of one chain
            walker_id (int): ID of the chain.
            n_steps (int): Number of steps to run.
            verbose (bool): Whether to print progress.

        Returns:
            tuple[int, np.ndarray, np.random.Generator]: A tuple of the
            walker ID, the log_thetas sampled by this walker, and the
            random number generator state after sampling.
        """
        prev_n = self.log_thetas.shape[1]

        kernel = self.kernel_class(
            rng=self.kernel_rngs[walker_id],
            step_size=self.step_size if isinstance(
                self.step_size, float) else self.step_size[walker_id],
            grad_and_log_likelihood=self.grad_and_log_likelihood,
            log_prior=self.log_prior,
            shape=self.shape,
            **{arg: getattr(self, arg)
                for arg in self._kernel_args[self.kernel_class]},
        )

        log_thetas = np.empty((n_steps // self.thin, self.size))

        prev_step_res = kernel.get_params(prev_step)

        for r in range(n_steps):
            if verbose and walker_id == 0:
                print(
                    f"Step {prev_n * self.thin + r + 1:6}/{prev_n * self.thin + n_steps:6}", end="\r")

            prev_step, prev_step_res, _, _ = kernel.one_step(
                prev_step, prev_step_res, return_info=True
            )

            if (r + 1) % self.thin == 0:
                log_thetas[r // self.thin] = prev_step

        return walker_id, log_thetas, kernel.rng

    def _run(self, n_steps: int, verbose: bool):
        """Internal run function. Calls walker function in parallel.

        Args:
            n_steps (int): Number of steps to run.
            verbose (bool): Whether to print progress.
        """
        with mp.Pool() as pool:
            results = pool.starmap(
                self._walker,
                [(
                    self.log_thetas[i, -1, :] if self.log_thetas.shape[1] > 0
                    else self.initial_step[i, 0, :],
                    i, n_steps, verbose)
                 for i in range(self.n_chains)],
            )

        # Update kernel_rngs
        self.log_thetas = np.concatenate(
            [self.log_thetas, np.empty(
                (self.n_chains, n_steps // self.thin, self.size))],
            axis=1
        )

        for walker_id, log_thetas, kernel_rng in results:
            self.kernel_rngs[walker_id] = kernel_rng
            if n_steps // self.thin > 0:
                self.log_thetas[walker_id, -
                                (n_steps // self.thin):] = log_thetas

    def run(self,
            stopping_crit: Literal["r_hat", "ESS"] | Callable | None = "r_hat",
            max_steps: int | None = None, check_interval: int | None = 1000,
            burn_in: int | float = 0.2, verbose: bool = True) -> np.ndarray:
        """Run MCMC sampling until the stopping criterion is met or the
        maximum number of steps is reached.

        Args:
            stopping_crit (Literal["r_hat", "ESS"] | Callable | None, optional):
                The stopping criterion to use. If ``"r_hat"``, runs
                until the the Gelman-Rubin potential scale reduction
                factor :math:`\\hat R` is below 1.01. If ``"ESS"``, runs
                until the effective sample size (ESS) is above 100. If a
                callable, it should take in the log_thetas and return a
                boolean indicating whether to stop. If ``None``, runs
                until ``max_steps`` is reached. Burn-in is discarded
                before checking the stopping criterion and only every
                ``check_interval`` steps. Defaults to ``"r_hat"``.
            max_steps (int | None, optional): Maximum number of steps to
                run. If ``None``, runs a maximum of 1,000,000 steps for
                RWM and MALA kernels and 100,000 steps for smMALA
                kernels. Defaults to None.
            check_interval (int | None, optional): Number of steps
                between checking the stopping criterion. Defaults to
                1000.
            burn_in (int | float, optional): Number of steps to discard
                as burn-in. If a float, it is interpreted as a fraction
                of the total steps. Defaults to 0.2.
            verbose (bool, optional): Whether to print progress.
                Defaults to True.

        Returns:
            np.ndarray: The log-thetas for each chain.
        """
        if stopping_crit == "r_hat":
            def stopping_crit(log_thetas):
                if log_thetas.shape[1] < self.n_chains:
                    return False
                return np.all(
                    np.array(arviz.rhat(arviz.convert_to_dataset(log_thetas)
                                        ).to_array()) < 1.01)

        elif stopping_crit == "ESS":
            def stopping_crit(log_thetas):
                if log_thetas.shape[1] < self.n_chains:
                    return False
                return np.all(
                    np.array(arviz.ess(arviz.convert_to_dataset(log_thetas)
                                       ).to_array()) > 100)

        elif stopping_crit is None:
            def stopping_crit(log_thetas): return False

        max_steps = max_steps or (
            1_000_000 if self.kernel_class in [RWMKernel, MALAKernel]
            else 100_000 if self.kernel_class == smMALAKernel
            else None)

        max_steps = (max_steps // self.thin) * self.thin

        if isinstance(self.step_size, str) and self.step_size == "auto":
            if verbose:
                print("Tuning step size...")
            self.tune_stepsize(verbose=verbose)
            if verbose:
                print(f"Using step size: {self.step_size}")

        if self.log_thetas.shape[1] == 0:
            self._take_initial_step()

        if max_steps == 0:
            return self.log_thetas

        while max_steps and not stopping_crit(
            self.log_thetas[:, burn_in if isinstance(burn_in, int)
                            else int(burn_in * self.log_thetas.shape[1]):, :]):
            self._run(min(check_interval, max_steps),
                      verbose=verbose)
            max_steps -= min(check_interval, max_steps)

        return self.log_thetas

    def acceptance(
            self, burn_in: int | float = 0.2, chain_id: int | None = None
    ) -> np.ndarray | float:
        """Calculate the acceptance rate for the chains.

        Args:
            burn_in (int | float, optional): Number of steps to discard
                as burn-in. If a float, it is interpreted as a fraction
                of the total steps. Defaults to 0.2.
            chain_id (int | None, optional): Chain ID to return
                acceptance rate for. If None, returns acceptance rates
                for all chains. Defaults to None.

        Returns:
            np.ndarray | float: The acceptance rate(s) for the specified
            chain(s).
        """
        if isinstance(burn_in, float):
            burn_in = int(burn_in * self.log_thetas.shape[1])

        log_thetas = self.log_thetas[:, burn_in:, :]

        acceptance_rates = list()

        for i in range(chain_id or log_thetas.shape[0]):
            accepted = np.sum(
                np.any(
                    log_thetas[i, 1:, :] != log_thetas[i, :-1, :],
                    axis=-1,
                )
            )
            total = log_thetas.shape[1] - 1
            acceptance_rates.append(accepted / total)

        return np.array(acceptance_rates) if chain_id is None \
            else acceptance_rates[0]

    def __getstate__(self):
        sampler = self.__dict__.copy()
        sampler.pop("grad_and_log_likelihood")
        sampler.pop("log_prior")
        sampler.pop("log_prior_grad", None)
        sampler.pop("log_prior_hessian", None)
        sampler.pop("data", None)

        return sampler

    def __setstate__(self, sampler):

        self.__dict__.update(sampler)
        self.grad_and_log_likelihood = self._get_grad_and_log_likelihood()
        log_prior = self._log_prior or (None, None, None)
        self.log_prior = log_prior[0] or self._get_log_prior(
            self._penalty[0])
        if self.kernel_class in [MALAKernel, smMALAKernel]:
            self.log_prior_grad = log_prior[1] or self._get_log_prior_grad(
                self._penalty[1])
        if self.kernel_class == smMALAKernel:
            self.log_prior_hessian = log_prior[2] or \
                self._get_log_prior_hessian(self._penalty[2])
        self.data = StateContainer(sampler["_bin_datamatrix"])

    def tune_stepsize(self, n_steps: int = 100, burn_in: float | int = 0.6,
                      target_acceptance: float | Literal["auto"] = "auto",
                      max_trials: int = 10, verbose: bool = True,
                      tol: float = 0.02,
                      ) -> float:
        """Automatically infer an appropriate step size epsilon for MCMC
        sampling.

        Args:
            n_steps (int, optional): Number of steps to run for
                inference. Defaults to 100.
            burn_in (float | int, optional): Burn-in period. If float,
                fraction of n_steps. If int, number of steps. Defaults
                to 0.6.
            target_acceptance (float | Literal["auto"], optional):
                Target acceptance rate. If "auto", set to 0.234 for RWM
                kernels, 0.574 for MALA kernels and 0.7 for smMALA
                kernels. Defaults to "auto".
            max_trials (int, optional): Maximum number of trials to run.
                Defaults to 10.
            verbose (bool, optional): Whether to print progress.
                Defaults to True.
            tol (float, optional): Tolerance for acceptance rate. If the
                acceptance rate is within ``tol`` of the target
                acceptance rate, the step size is accepted. Defaults to
                0.02.

        Returns:
            float: Inferred step sizet.
        """
        if target_acceptance == "auto":
            if self.kernel_class == RWMKernel:
                target_acceptance = 0.234
            elif self.kernel_class == MALAKernel:
                target_acceptance = 0.574
            elif self.kernel_class == smMALAKernel:
                target_acceptance = 0.7

        n_parallel = 5
        step_sizes = 10 ** np.linspace(-5, -1, n_parallel)

        for trial in range(max_trials):
            if verbose:
                print(f"Trial {trial+1}: step_sizes={step_sizes}")

            model = (oMHN if self.omhn else cMHN)(
                log_theta=np.zeros(self.shape),
                meta={"lambda": self.lam / self.n_samples},
            )
            temp_sampler = MCMC(
                mhn_model=model,
                data=self._bin_datamatrix,
                n_chains=n_parallel * 3,
                step_size=step_sizes.repeat(3),
                penalty=self._penalty,
                log_prior=self._log_prior,
                kernel_class=self.kernel_class,
                thin=1,
                seed=self.rng.integers(0, 2**32, dtype=np.uint32)
            )

            temp_sampler._penalty = self._penalty

            if self.initial_step is not None:
                temp_sampler.initial_step = np.stack(
                    [self.initial_step[i % self.n_chains, :, :]
                     for i in range(n_parallel * 3)])

            try:
                temp_sampler.run(max_steps=n_steps, verbose=verbose,
                                 stopping_crit=None)
            except ZeroDivisionError as e:
                if verbose:
                    print(f"Trial {trial+1} failed with error: {e}\n"
                          "Decreasing step sizes.")
                step_sizes *= 0.1
                continue

            acceptance_rates = temp_sampler.acceptance(
                burn_in=burn_in).reshape(n_parallel, 3).mean(axis=1)

            if verbose:
                print(f"Acceptance rates: {acceptance_rates}")

            argbest = np.argmin(np.abs(acceptance_rates - target_acceptance))
            if np.abs(acceptance_rates[argbest] - target_acceptance) < tol:
                self.step_size = step_sizes[argbest]
                return step_sizes[argbest]
            if acceptance_rates[argbest] < target_acceptance:
                step_sizes = np.linspace(
                    (step_sizes[argbest - 1] if argbest > 0
                     else step_sizes[argbest] / 10),
                    step_sizes[argbest],
                    n_parallel)
            else:
                step_sizes = np.linspace(
                    step_sizes[argbest],
                    (step_sizes[argbest + 1] if argbest < n_parallel - 1
                     else step_sizes[argbest] * 10),
                    n_parallel)

    def rhat(self, burn_in: int | float = 0.2, **kwargs) -> np.ndarray:
        """Calculate the Gelman-Rubin potential scale reduction factor
        :math:`\\hat R`.

        Args:
            burn_in (int | float, optional): Number of steps to discard
                as burn-in. If a float, it is interpreted as a fraction
                of the total steps. Defaults to 0.2.
            **kwargs: Additional keyword arguments to pass to
                :func:`arviz.rhat`. See the ArviZ documentation for more
                details.

        Returns:
            np.ndarray: The Gelman-Rubin R-hat values for each parameter.
        """
        if isinstance(burn_in, float):
            burn_in = int(burn_in * self.log_thetas.shape[1])

        log_thetas = self.log_thetas[:, burn_in:, :]

        return np.array(arviz.rhat(
            arviz.convert_to_inference_data(log_thetas), **kwargs).x)

    def ess(self, burn_in: int | float = 0.2, **kwargs):
        """Calculate the effective sample size (ESS).

        Args:
            burn_in (int | float, optional): Number of steps to discard
                as burn-in. If a float, it is interpreted as a fraction
                of the total steps. Defaults to 0.2.
            **kwargs: Additional keyword arguments to pass to
                :func:`arviz.ess`. See the ArviZ documentation for more
                details.

        Returns:
            np.ndarray: The effective sample size for each parameter.
        """
        if isinstance(burn_in, float):
            burn_in = int(burn_in * self.log_thetas.shape[1])

        log_thetas = self.log_thetas[:, burn_in:, :]

        return np.array(arviz.ess(
            arviz.convert_to_inference_data(log_thetas), **kwargs).x)



"""
Unit tests for the mhn.mcmc module.
"""

import sys
import unittest
if sys.version_info < (3, 10):
    raise unittest.SkipTest("mhn.mcmc requires Python 3.9+")

import numpy as np
import mhn
from mhn.training.state_containers import StateContainer
from mhn.mcmc.kernels import RWMKernel, MALAKernel, smMALAKernel
from mhn.mcmc.mcmc import MCMC
from mhn.optimizers import Penalty
import scipy.stats

np.random.seed(0)
data = np.random.randint(2, size=(100,3), dtype=np.int32)
data_container = StateContainer(data)
data_size = data.shape[0]
optimizer = mhn.optimizers.oMHNOptimizer().load_data_matrix(data)
np.random.seed(0)
lam = optimizer.lambda_from_cv()
model = optimizer.train(lam=lam)
shape = (4, 3)
size = 12
lam = 0.001


def grad_and_log_likelihood(theta):
    grad, lik = mhn.training.likelihood_omhn.gradient_and_score(
        theta.reshape(shape), data_container)
    return data_size * grad, data_size * lik


def log_l2_prior(theta):
    return -lam * np.sum(theta**2)


def log_l2_prior_grad(theta):
    return -2 * lam * theta.reshape(shape)


def log_l2_prior_hessian(theta):
    return -2 * lam * np.eye(size)


def log_l1_prior(theta):
    return -lam * np.sum(np.abs(theta))


def log_l1_prior_grad(theta):
    return -lam * np.sign(theta.reshape(shape))


def log_l1_prior_hessian(theta):
    return np.zeros((size, size))


initial_theta = np.random.normal(loc=0, scale=1 / np.sqrt(2 * lam), size=size)


class TestKernels(unittest.TestCase):
    """Tests for MCMC kernel classes."""

    def test_rwm_kernel_init(self):
        """Test RWMKernel initialization."""
        kernel = RWMKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            step_size=0.1,
            shape=shape,
        )
        self.assertEqual(kernel.step_size, 0.1)
        self.assertEqual(kernel.shape, shape)
        self.assertEqual(kernel.size, size)

    def test_rwm_kernel_get_params(self):
        """Test RWMKernel get_params method."""
        kernel = RWMKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            step_size=0.1,
            shape=shape,
        )

        result = kernel.get_params(initial_theta)

        assert (np.allclose(result.log_likelihood,
                grad_and_log_likelihood(initial_theta)[1]))
        assert (np.allclose(result.log_prior, log_l2_prior(initial_theta)))

    def test_rwm_kernel_propose(self):
        """Test RWMKernel propose method."""
        kernel = RWMKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(42))
        )

        prev_step_res = kernel.get_params(initial_theta)
        kernel.propose(
            initial_theta, prev_step_res)

    def test_rwm_kernel_log_accept(self):
        """Test RWMKernel log_accept method."""
        kernel = RWMKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            step_size=0.1,
            shape=shape,
        )

        step1 = initial_theta
        res1 = kernel.get_params(step1)
        step2 = initial_theta + 0.1
        res2 = kernel.get_params(step2)

        accept_ratio = kernel.log_accept(step1, res1, step2, res2)

        assert (np.allclose(
            accept_ratio,
            log_l2_prior(step2)
            - log_l2_prior(step1)
            + grad_and_log_likelihood(step2)[1]
            - grad_and_log_likelihood(step1)[1]))

    def test_rwm_kernel_one_step(self):
        """Test RWMKernel one_step method."""
        kernel = RWMKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(42))
        )

        prev_step_res = kernel.get_params(initial_theta)
        new_step, new_step_res, acceptance_ratio, accepted = kernel.one_step(
            initial_theta, prev_step_res, return_info=True
        )

        # Check output types and shapes
        self.assertEqual(new_step.shape, initial_theta.shape)
        self.assertIsInstance(acceptance_ratio, float)
        self.assertIn(accepted, [0, 1])

    def test_rwm_kernel_seed_reproducibility(self):
        """Test RWMKernel reproducibility with same seed."""
        # First run with seed 42
        kernel1 = RWMKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(42))
        )

        prev_step_res1 = kernel1.get_params(initial_theta)
        new_step1, _ = kernel1.one_step(initial_theta, prev_step_res1)

        # Second run with same seed 42
        kernel2 = RWMKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(42))
        )

        prev_step_res2 = kernel2.get_params(initial_theta)
        new_step2, _ = kernel2.one_step(initial_theta, prev_step_res2)

        # Results should be identical
        np.testing.assert_array_equal(new_step1, new_step2)

    def test_rwm_kernel_different_seeds_differ(self):
        """Test that RWMKernel with different seeds produces different results."""
        # First run with seed 42
        kernel1 = RWMKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(42))
        )

        prev_step_res1 = kernel1.get_params(initial_theta)
        new_step1, _ = kernel1.propose(initial_theta, prev_step_res1)

        # Second run with different seed 99
        kernel2 = RWMKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(99))
        )

        prev_step_res2 = kernel2.get_params(initial_theta)
        new_step2, _ = kernel2.propose(initial_theta, prev_step_res2)

        # Results should be different (with high probability)
        self.assertFalse(np.allclose(new_step1, new_step2))

    def test_mala_kernel_init(self):
        """Test MALAKernel initialization."""
        kernel = MALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            step_size=0.1,
            shape=shape,
        )
        self.assertEqual(kernel.step_size, 0.1)
        self.assertEqual(kernel.shape, shape)

    def test_mala_kernel_get_params(self):
        """Test MALAKernel get_params method."""
        kernel = MALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            step_size=0.1,
            shape=shape,
        )

        result = kernel.get_params(initial_theta)

        assert (np.allclose(result.log_likelihood,
                grad_and_log_likelihood(initial_theta)[1]))
        assert (np.allclose(result.log_prior, log_l2_prior(initial_theta)))
        self.assertTrue(hasattr(result, "mu"))

    def test_mala_kernel_propose(self):
        """Test MALAKernel propose method."""
        kernel = MALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(42))
        )

        prev_step_res = kernel.get_params(initial_theta)
        kernel.propose(
            initial_theta, prev_step_res)

    def test_mala_kernel_log_accept(self):
        """Test MALAKernel log_accept method."""
        kernel = MALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            step_size=0.1,
            shape=shape,
        )

        step1 = initial_theta
        res1 = kernel.get_params(step1)
        step2 = initial_theta + 0.1
        res2 = kernel.get_params(step2)

        accept_ratio = kernel.log_accept(step1, res1, step2, res2)

        assert (np.allclose(
            accept_ratio,
            log_l2_prior(step2)
            - log_l2_prior(step1)
            + grad_and_log_likelihood(step2)[1]
            - grad_and_log_likelihood(step1)[1]
            + scipy.stats.multivariate_normal.logpdf(
                step1,
                mean=step2 + kernel.step_size / 2 *
                (grad_and_log_likelihood(step2)[
                 0] + log_l2_prior_grad(step2)).flatten(),
                cov=kernel.step_size * np.eye(size))
            - scipy.stats.multivariate_normal.logpdf(
                step2,
                mean=step1 + kernel.step_size / 2 *
                (grad_and_log_likelihood(step1)[
                 0] + log_l2_prior_grad(step1)).flatten(),
                cov=kernel.step_size * np.eye(size))
        ))

    def test_mala_kernel_one_step(self):
        """Test MALAKernel one_step method."""
        kernel = MALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(42))
        )

        prev_step_res = kernel.get_params(initial_theta)
        new_step, new_step_res, acceptance_ratio, accepted = kernel.one_step(
            initial_theta, prev_step_res, return_info=True
        )

        # Check output types and shapes
        self.assertEqual(new_step.shape, initial_theta.shape)
        self.assertIsInstance(acceptance_ratio, (float, np.floating))
        self.assertIn(accepted, [0, 1])

    def test_mala_kernel_seed_reproducibility(self):
        """Test MALAKernel reproducibility with same seed."""
        # First run with seed 123
        kernel1 = MALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(123))
        )

        prev_step_res1 = kernel1.get_params(initial_theta)
        new_step1, _ = kernel1.one_step(initial_theta, prev_step_res1)

        # Second run with same seed 123
        kernel2 = MALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(123))
        )

        prev_step_res2 = kernel2.get_params(initial_theta)
        new_step2, _ = kernel2.one_step(initial_theta, prev_step_res2)

        # Results should be identical
        np.testing.assert_array_equal(new_step1, new_step2)

    def test_mala_kernel_different_seeds_differ(self):
        """Test that MALAKernel with different seeds produces different results."""
        # First run with seed 123
        kernel1 = MALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(123))
        )

        prev_step_res1 = kernel1.get_params(initial_theta)
        new_step1, _ = kernel1.propose(initial_theta, prev_step_res1)

        # Second run with different seed 321
        kernel2 = MALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            step_size=0.1,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(321))
        )

        prev_step_res2 = kernel2.get_params(initial_theta)
        new_step2, _ = kernel2.propose(initial_theta, prev_step_res2)

        # Results should be different (with high probability)
        self.assertFalse(np.allclose(new_step1, new_step2))

    def test_smmala_kernel_init(self):
        """Test smMALAKernel initialization."""
        kernel = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.1,
            shape=shape,
        )
        self.assertEqual(kernel.step_size, 0.1)
        self.assertEqual(kernel.shape, shape)
        self.assertFalse(kernel.use_cuda)

    def test_smmala_kernel_get_params(self):
        """Test smMALAKernel get_params method."""
        kernel = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.1,
            shape=shape,
        )

        result = kernel.get_params(initial_theta)

        assert (np.allclose(result.log_likelihood,
                grad_and_log_likelihood(initial_theta)[1]))
        assert (np.allclose(result.log_prior, log_l2_prior(initial_theta)))
        assert (np.allclose(result.gradient, grad_and_log_likelihood(
            initial_theta)[0] + log_l2_prior_grad(initial_theta)))
        assert (np.allclose(result.G, -log_l2_prior_hessian(initial_theta) +
                mhn.full_state_space.fisher.fisher(initial_theta.reshape(shape))))
        assert (np.allclose(result.cholesky, np.linalg.cholesky(result.G)))
        assert hasattr(result, 'mu')
        assert (np.allclose(result.det_sqrt, np.sqrt(np.linalg.det(result.G))))

    def test_smmala_kernel_get_params_cuda(self):
        """Test smMALAKernel get_params method with CUDA."""
        if mhn.cuda_available() != mhn.CUDA_AVAILABLE:
            self.skipTest("CUDA not available, skipping CUDA test.")
        kernel = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.1,
            shape=shape,
        )
        cuda_kernel = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.1,
            shape=shape,
            use_cuda=True
        )

        result = kernel.get_params(initial_theta)
        result_cuda = cuda_kernel.get_params(initial_theta)

        # Check that results are close (accounting for possible floating point differences)
        self.assertTrue(np.allclose(result.log_likelihood,
                        result_cuda.log_likelihood))
        self.assertTrue(np.allclose(result.log_prior, result_cuda.log_prior))
        self.assertTrue(np.allclose(result.gradient, result_cuda.gradient))
        self.assertTrue(np.allclose(result.G, result_cuda.G))
        self.assertTrue(np.allclose(result.cholesky, result_cuda.cholesky))
        self.assertTrue(np.allclose(result.mu, result_cuda.mu))
        self.assertTrue(np.allclose(result.det_sqrt, result_cuda.det_sqrt))

    def test_linalg_error_non_positive_definite(self):
        """Test that smMALAKernel raises LinAlgError for non-positive-definite metric."""

        kernel = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l1_prior,
            log_prior_grad=log_l1_prior_grad,
            log_prior_hessian=log_l1_prior_hessian,
            step_size=0.1,
            shape=shape,
        )

        with self.assertRaises(np.linalg.LinAlgError):
            kernel.get_params(initial_theta)

    def test_smmala_kernel_propose(self):
        """Test smMALAKernel propose method."""
        kernel = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.01,  # Smaller step size for stability
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(42))
        )

        prev_step_res = kernel.get_params(initial_theta)
        kernel.propose(
            initial_theta, prev_step_res)

    def test_smmala_kernel_log_accept(self):
        """Test smMALAKernel log_accept method."""
        kernel = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.01,
            shape=shape,
        )

        step1 = initial_theta
        res1 = kernel.get_params(step1)
        step2 = initial_theta + 0.001
        res2 = kernel.get_params(step2)

        accept_ratio = kernel.log_accept(step1, res1, step2, res2)

        assert (np.allclose(
            accept_ratio,
            log_l2_prior(step2)
            - log_l2_prior(step1)
            + grad_and_log_likelihood(step2)[1]
            - grad_and_log_likelihood(step1)[1]
            + scipy.stats.multivariate_normal.logpdf(
                step1,
                mean=step2 + kernel.step_size / 2 * np.linalg.inv(res2.G) @
                (grad_and_log_likelihood(step2)[
                 0] + log_l2_prior_grad(step2)).flatten(),
                cov=kernel.step_size * np.linalg.inv(res2.G))
            - scipy.stats.multivariate_normal.logpdf(
                step2,
                mean=step1 + kernel.step_size / 2 * np.linalg.inv(res1.G) @
                (grad_and_log_likelihood(step1)[
                 0] + log_l2_prior_grad(step1)).flatten(),
                cov=kernel.step_size * np.linalg.inv(res1.G))
        ))

    def test_smmala_kernel_one_step(self):
        """Test smMALAKernel one_step method."""
        kernel = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.001,  # Very small step size for stability
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(42))
        )
        prev_step_res = kernel.get_params(initial_theta)
        new_step, new_step_res, acceptance_ratio, accepted = kernel.one_step(
            initial_theta, prev_step_res, return_info=True
        )

        # Check output types and shapes
        self.assertEqual(new_step.shape, initial_theta.shape)
        self.assertIsInstance(acceptance_ratio, (float, np.floating))
        self.assertIn(accepted, [0, 1])

    def test_smmala_kernel_seed_reproducibility(self):
        """Test smMALAKernel reproducibility with same seed."""
        # First run with seed 456
        kernel1 = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.001,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(456))
        )

        prev_step_res1 = kernel1.get_params(initial_theta)
        new_step1, _ = kernel1.one_step(initial_theta, prev_step_res1)

        # Second run with same seed 456
        kernel2 = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.001,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(456))
        )

        prev_step_res2 = kernel2.get_params(initial_theta)
        new_step2, _ = kernel2.one_step(initial_theta, prev_step_res2)

        # Results should be identical
        np.testing.assert_array_equal(new_step1, new_step2)

    def test_smmala_kernel_different_seeds_differ(self):
        """Test that smMALAKernel with different seeds produces different results."""
        # First run with seed 456
        kernel1 = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.001,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(456))
        )

        prev_step_res1 = kernel1.get_params(initial_theta)
        new_step1, _ = kernel1.propose(initial_theta, prev_step_res1)

        # Second run with different seed 654
        kernel2 = smMALAKernel(
            grad_and_log_likelihood=grad_and_log_likelihood,
            log_prior=log_l2_prior,
            log_prior_grad=log_l2_prior_grad,
            log_prior_hessian=log_l2_prior_hessian,
            step_size=0.001,
            shape=shape,
            rng=np.random.Generator(np.random.PCG64(654))
        )

        prev_step_res2 = kernel2.get_params(initial_theta)
        new_step2, _ = kernel2.propose(initial_theta, prev_step_res2)

        # Results should be different (with high probability)
        self.assertFalse(np.allclose(new_step1, new_step2))


class TestMCMC(unittest.TestCase):
    """Integration tests for the MCMC sampler."""

    def test_init_from_optimizer(self):
        sampler = MCMC(optimizer=optimizer, n_chains=2)
        self.assertEqual(sampler.n_chains, 2)

    def test_init_from_model_data_penalty(self):
        """Initialize using model, data and penalty from optimizer."""
        sampler = MCMC(
            mhn_model=model,
            data=data,
            penalty=Penalty.L2,
            n_chains=2
        )
        self.assertEqual(sampler.n_chains, 2)

    def test_init_invalid_combinations(self):
        """Check errors when providing conflicting or incomplete arguments."""
        params = {
            "optimizer": optimizer,
            "data": data,
            "mhn_model": model,
            "penalty": Penalty.L2,
            "log_prior": (log_l2_prior, log_l2_prior_grad, log_l2_prior_hessian),
        }

        for args in [
            ("data",),
            ("mhn_model",),
            ("penalty",),
            ("log_prior",),
            ("data", "mhn_model"),
            ("data", "penalty"),
            ("data", "log_prior"),
            ("mhn_model", "penalty"),
            ("mhn_model", "log_prior"),
            ("penalty", "log_prior"),
            ("data", "mhn_model", "penalty", "log_prior"),

        ]:
            with self.assertRaises(ValueError):
                MCMC(**{k: params[k] for k in args})

        with self.assertRaises(ValueError):
            optimizer2 = mhn.optimizers.oMHNOptimizer()
            optimizer2.results = optimizer.result
            optimizer2._data = optimizer._data
            MCMC(optimizer=optimizer2)
        with self.assertWarns(UserWarning):
            MCMC(optimizer=optimizer, penalty=Penalty.L2)
        with self.assertWarns(UserWarning):
            MCMC(optimizer=optimizer, log_prior=(
                log_l2_prior, log_l2_prior_grad, log_l2_prior_hessian))
        with self.assertRaises(ValueError):
            MCMC(optimizer=optimizer, kernel_class=smMALAKernel)
        with self.assertRaises(ValueError):
            MCMC(mhn_model=model, log_prior=(log_l2_prior, log_l2_prior_grad),
                 kernel_class=smMALAKernel)
        with self.assertRaises(ValueError):
            MCMC(mhn_model=model, log_prior=(log_l2_prior),
                 kernel_class=MALAKernel)

    def test_run_rwm_basic(self):
        sampler = MCMC(
            optimizer=optimizer,
            kernel_class=RWMKernel,
            step_size=0.1,
            n_chains=2,
            thin=1,
            seed=42
        )
        result = sampler.run(stopping_crit=None, max_steps=10, verbose=False)
        self.assertEqual(result.shape, (2, 10, size))

        sampler = MCMC(
            mhn_model=model,
            data=data,
            penalty=Penalty.L2,
            kernel_class=RWMKernel,
            step_size=0.1,
            n_chains=2,
            thin=1,
            seed=42
        )
        result = sampler.run(stopping_crit=None, max_steps=10, verbose=False)
        self.assertEqual(result.shape, (2, 10, size))

    def test_run_mala_basic(self):
        sampler = MCMC(
            optimizer=optimizer,
            kernel_class=MALAKernel,
            step_size=0.1,
            n_chains=2,
            thin=1,
            seed=42
        )
        result = sampler.run(stopping_crit=None, max_steps=10, verbose=False)
        self.assertEqual(result.shape, (2, 10, size))

    def test_run_smmala_basic(self):
        sampler = MCMC(
            mhn_model=model,
            data=data,
            log_prior=(log_l2_prior, log_l2_prior_grad, log_l2_prior_hessian),
            kernel_class=smMALAKernel,
            step_size=0.001,
            n_chains=2,
            thin=1,
            seed=42
        )
        sampler.initial_step = np.random.normal(
            loc=0, scale=1 / np.sqrt(2 * lam), size=(2, 1, size))
        result = sampler.run(stopping_crit=None,
                             max_steps=3, verbose=False)
        self.assertEqual(result.shape, (2, 3, size))

    def test_seed_reproducibility(self):
        sam1 = MCMC(
            optimizer=optimizer,
            step_size=0.1,
            seed=123,
            n_chains=2,
            thin=1
        )
        sam1.run(stopping_crit=None, max_steps=5, verbose=False)
        sam2 = MCMC(
            optimizer=optimizer,
            step_size=0.1,
            seed=123,
            n_chains=2,
            thin=1
        )
        sam2.run(stopping_crit=None, max_steps=5, verbose=False)
        np.testing.assert_array_equal(sam1.log_thetas, sam2.log_thetas)

    def test_rhat_and_ess(self):
        sampler = MCMC(
            optimizer=optimizer,
            step_size=0.1,
            seed=42,
            n_chains=3,
            thin=1
        )
        sampler.run(stopping_crit=None, max_steps=20, verbose=False)
        r = sampler.rhat()
        e = sampler.ess()
        self.assertEqual(len(r), sampler.size)
        self.assertEqual(len(e), sampler.size)
        self.assertTrue(np.all(r > 0.99))
        self.assertTrue(np.all(e > 0))

    def test_acceptance_rate(self):
        sampler = MCMC(
            optimizer=optimizer,
            step_size=0.1,
            seed=42,
            n_chains=3,
            thin=1
        )
        sampler.run(stopping_crit=None, max_steps=20, verbose=False)
        rates = sampler.acceptance()
        self.assertEqual(len(rates), 3)
        self.assertTrue(np.all((rates >= 0) & (rates <= 1)))

    def test_tune_stepsize(self):
        sampler = MCMC(
            optimizer=optimizer,
            step_size="auto",
            n_chains=2,
            seed=42,
            thin=1
        )
        val = sampler.tune_stepsize(n_steps=5, verbose=False, tol=0.3)
        self.assertIsInstance(val, float)
        self.assertEqual(sampler.step_size, val)


if __name__ == "__main__":
    unittest.main()



"""
This module contains some utility functions for working with MHNs.
"""
# author(s): Stefan Vocht

cimport cython
from libc.math cimport exp

import numpy as np
cimport numpy as np


np.import_array()


def set_seed(seed: int):
    """
    Set the random seed for reproducibility. Internally, this sets the seed for the numpy random generator.

    Parameters:
        seed (int): The seed value to use for random number generators.
    """
    # if the package should ever use other sources of randomness in the future, add the corresponding seed setting functions here
    np.random.seed(seed)


@cython.wraparound(False)
@cython.boundscheck(False)
def sample_artificial_data(np.ndarray[np.double_t, ndim=2] theta, int sample_num) -> np.ndarray:
    """
    Samples artificial cross-sectional data from a cMHN. Use np.random.seed() to make results reproducible.

    Args:
        theta (np.ndarray[np.double_t, ndim=2]): Theta matrix representing a cMHN.
        sample_num (int): Number of samples in the generated data.

    Returns:
        np.ndarray: 2D array where each row corresponds to a sample and each column to a gene.
    """
    cdef int n = theta.shape[0]
    cdef np.ndarray[np.double_t, ndim=2] exp_theta = np.exp(theta)
    cdef np.ndarray[np.int32_t, ndim=2] art_data = np.zeros((sample_num, n), dtype=np.int32)

    cdef np.ndarray[np.int32_t] in_current_sample = np.zeros(n, dtype=np.int32)
    cdef np.ndarray[np.int32_t] possible_gene_mutations = np.empty(n, dtype=np.int32)
    cdef np.ndarray[np.double_t] rates_from_current_state = np.empty(n, dtype=np.double)

    cdef int j, gene, mutated_gene
    cdef int sample_index, current_sample_length
    cdef double observation_time, current_time, passed_time
    cdef double sum_rates, rate, random_crit, accumulated_rate

    for sample_index in range(sample_num):
        in_current_sample[:] = 0
        current_sample_length = 0
        observation_time = np.random.exponential(1)
        current_time = 0.
        while current_sample_length < n:
            sum_rates = 0.
            j = 0
            for gene in range(n):
                if not in_current_sample[gene]:
                    possible_gene_mutations[j] = gene
                    rate = exp_theta[gene, gene]
                    for mutated_gene in range(n):
                        if in_current_sample[mutated_gene]:
                            rate *= exp_theta[gene, mutated_gene]
                    rates_from_current_state[j] = rate
                    sum_rates += rate
                    j += 1

            passed_time = np.random.exponential(1/sum_rates)
            current_time += passed_time
            if current_time > observation_time:
                break
            random_crit = np.random.random(1)[0] * sum_rates
            accumulated_rate = 0.
            for j in range(n - current_sample_length):
                rate = rates_from_current_state[j]
                gene = possible_gene_mutations[j]
                accumulated_rate += rate
                if random_crit <= accumulated_rate:
                    in_current_sample[gene] = 1
                    current_sample_length += 1
                    art_data[sample_index, gene] = 1
                    break
    return art_data


@cython.wraparound(False)
@cython.boundscheck(False)
def gillespie(np.ndarray[np.double_t, ndim=2] theta, np.ndarray[np.int32_t, ndim=1] initial_state, int sample_num, bint return_event_times) -> tuple[list[list[int]], np.ndarray | list[list[int]], list[list[float]], np.ndarray]:
    """
    Simulates event accumulation using the Gillespie algorithm.

    Args:
        theta (np.ndarray[np.double_t, ndim=2]): Theta matrix representing a cMHN or an oMHN.
        initial_state (np.ndarray[np.int32_t, ndim=1]): Array representing the starting state.
            Each entry corresponds to an event being present (1) or not (0).
        sample_num (int): Number of samples to simulate.
        return_event_times (bint): If True, returns the times at which all events occurred.

    Returns:
        tuple[list[list[int]], np.ndarray | list[list[int]], list[list[float]], np.ndarray]: A tuple containing:
            - A list of lists where each inner list contains active events in chronological order.
            - if return_event_times is True, a list of lists of timepoints at which the events occurred for every sample (occurrence times of events already present in initial_state are declared as None),
            - A numpy array of observation times.

    Raises:
        ValueError: If the size of theta is neither (n, n) nor (n+1, n).
    """
    cdef int n = theta.shape[1]
    cdef np.ndarray[np.double_t, ndim=2] exp_theta = np.exp(theta)
    cdef np.ndarray[np.double_t, ndim=1] observation_times = np.empty(sample_num, dtype=np.double)

    cdef list trajectory_list = []
    cdef list trajectory

    cdef list times_list = []
    cdef list times

    cdef np.ndarray[np.int32_t] in_current_sample = np.zeros(n, dtype=np.int32)
    cdef np.ndarray[np.int32_t] possible_gene_mutations = np.empty(n, dtype=np.int32)
    cdef np.ndarray[np.double_t] rates_from_current_state = np.empty(n, dtype=np.double)
    cdef np.ndarray[np.double_t] observation_rates

    if theta.shape[0] == n:
        # cMHN
        observation_rates = np.ones(n, dtype=np.double)
    elif theta.shape[0] == n+1:
        # oMHN
        observation_rates = exp_theta[n]
    else:
        raise ValueError(f"Theta must have dimensions {n}x{n} or {n+1}x{n}, but is has dimension {theta.shape[0]}x{theta.shape[1]}")


    cdef int j, gene, mutated_gene
    cdef int sample_index, current_sample_length
    cdef double current_time, passed_time
    cdef double sum_rates, rate, random_crit, accumulated_rate

    cdef int initial_sample_length = initial_state.sum()

    for sample_index in range(sample_num):
        in_current_sample[:] = initial_state.copy()
        current_sample_length = initial_sample_length
        current_time = 0.
        trajectory = [gene for gene in range(n) if initial_state[gene]]
        if return_event_times:
            times = [None for gene in trajectory]

        while 1:
            sum_rates = 0.
            j = 0

            # compute observation rate first and add it to the total sum of rates that lead out of the current state
            rate = 1.
            for mutated_gene in range(n):
                if in_current_sample[mutated_gene]:
                    rate *= observation_rates[mutated_gene]
            sum_rates += rate

            # now compute all the rates that lead to other states
            for gene in range(n):
                if not in_current_sample[gene]:
                    possible_gene_mutations[j] = gene
                    rate = exp_theta[gene, gene]
                    for mutated_gene in range(n):
                        if in_current_sample[mutated_gene]:
                            rate *= exp_theta[gene, mutated_gene]
                    rates_from_current_state[j] = rate
                    sum_rates += rate
                    j += 1

            # sample holding time
            passed_time = np.random.exponential(1 / sum_rates)
            current_time += passed_time

            
            random_crit = np.random.random(1)[0] * sum_rates
            accumulated_rate = 0.
            for j in range(n - current_sample_length):
                rate = rates_from_current_state[j]
                gene = possible_gene_mutations[j]
                accumulated_rate += rate
                if random_crit <= accumulated_rate:
                    in_current_sample[gene] = 1
                    current_sample_length += 1
                    trajectory.append(gene)
                    if return_event_times:
                        times.append(current_time)
                    break
            else:
                # if we did not leave the loop early, this means no state was selected as next state
                # instead the observation event happened
                trajectory_list.append(trajectory)
                observation_times[sample_index] = current_time
                if return_event_times:
                    times_list.append(times)
                break

    return (trajectory_list, times_list, observation_times) if return_event_times else (trajectory_list, observation_times)


@cython.wraparound(False)
@cython.boundscheck(False)
def gillespie_timed(np.ndarray[np.double_t, ndim=2] theta, np.ndarray[np.int32_t, ndim=1] initial_state, int sample_num, float obs_time, bint return_times) -> tuple[list[list[int]] | list[list[int]], list[list[float]]]:
    """
    Simulates event accumulation using the Gillespie algorithm, but stops only after a specified observation time.

    Args:
        theta (np.ndarray[np.double_t, ndim=2]): Theta matrix representing a cMHN or an oMHN.
        initial_state (np.ndarray[np.int32_t, ndim=1]): Array representing the starting state.
            Each entry corresponds to an event being present (1) or not (0).
        sample_num (int): Number of samples to simulate.
        obs_time (float): The observation time after which the simulation stops. If np.inf, the simulation runs until all events have occurred.
        return_times (bint): If True, returns the times at which events occurred.

    Returns:
        tuple[list[list[int]] | list[list[int]], list[list[float]]]: A tuple containing:
            - A list of lists where each inner list contains active events in chronological order.
            - if return_times is True, a list of times at which events occurred (occurrence times of events already present in initial_state are declared as None).

    Raises:
        ValueError: If the size of theta is neither (n, n) nor (n+1, n).
    """
    cdef int n = theta.shape[1]
    cdef np.ndarray[np.double_t, ndim=2] exp_theta = np.exp(theta)

    cdef list trajectory_list = []
    cdef list trajectory

    cdef list times_list = []
    cdef list times

    cdef np.ndarray[np.int32_t] in_current_sample = np.zeros(n, dtype=np.int32)
    cdef np.ndarray[np.int32_t] possible_gene_mutations = np.empty(n, dtype=np.int32)
    cdef np.ndarray[np.double_t] rates_from_current_state = np.empty(n, dtype=np.double)

    cdef int j, gene, mutated_gene
    cdef int current_sample_length
    cdef double current_time, passed_time
    cdef double sum_rates, rate, random_crit, accumulated_rate

    cdef int initial_sample_length = initial_state.sum()

    for _ in range(sample_num):
        in_current_sample[:] = initial_state.copy()
        current_sample_length = initial_sample_length
        current_time = 0.
        trajectory = [gene for gene in range(n) if initial_state[gene]]
        if return_times:
            times = [None for gene in trajectory]

        while 1:
            sum_rates = 0.
            j = 0

            # compute all rates that lead to other states
            for gene in range(n):
                if not in_current_sample[gene]:
                    possible_gene_mutations[j] = gene
                    rate = exp_theta[gene, gene]
                    for mutated_gene in range(n):
                        if in_current_sample[mutated_gene]:
                            rate *= exp_theta[gene, mutated_gene]
                    rates_from_current_state[j] = rate
                    sum_rates += rate
                    j += 1

            if sum_rates == 0:
                # if no rates are left, we stop the simulation
                trajectory_list.append(trajectory)
                if return_times:
                    times_list.append(times)
                break

            # sample holding time
            passed_time = np.random.exponential(1 / sum_rates)
            current_time += passed_time
            if current_time > obs_time:
                # if we reached the observation time, we stop the simulation
                trajectory_list.append(trajectory)
                if return_times:
                    times_list.append(times)
                break

            random_crit = np.random.random(1)[0] * sum_rates
            accumulated_rate = 0.
            for j in range(n - current_sample_length):
                rate = rates_from_current_state[j]
                gene = possible_gene_mutations[j]
                accumulated_rate += rate
                if random_crit <= accumulated_rate:
                    in_current_sample[gene] = 1
                    current_sample_length += 1
                    trajectory.append(gene)
                    if return_times:
                        times.append(current_time)
                    break

    return (trajectory_list, times_list) if return_times else (trajectory_list,)


def compute_next_event_probs(np.ndarray[np.double_t, ndim=2] theta, np.ndarray[np.int32_t, ndim=1] current_state, double observation_rate = 0) -> np.ndarray:
    """
    Computes probabilities for each event to be the next one to occur, given the current state.

    Args:
        theta (np.ndarray[np.double_t, ndim=2]): Theta matrix representing a cMHN.
        current_state (np.ndarray[np.int32_t, ndim=1]): Array representing the current state.
            Each entry corresponds to an event being present (1) or not (0).
        observation_rate (float, optional): Rate of the observation event. Defaults to 0,
            meaning no observation occurs before another event.

    Returns:
        np.ndarray: Array containing the probabilities for each event to occur next.

    Raises:
        ValueError: If the size of theta does not match the size of current_state.
    """
    cdef int n = theta.shape[1]  # use shape[1] to be compatible with oMHN
    if n != current_state.shape[0]:
        raise ValueError(f"Number of events represented by theta ({n}) does not match the size of current_state ({current_state.shape[0]})")

    cdef np.ndarray[np.double_t, ndim=2] theta_copy = theta.copy()
    cdef np.ndarray[np.double_t, ndim=1] result = np.zeros(n, dtype=np.double)
    cdef int i, j
    cdef double rate

    for i in range(n):
        if current_state[i]:
            continue
        rate = theta[i, i]
        for j in range(n):
            if current_state[j]:
                rate += theta[i, j]
        result[i] = exp(rate)

    result /= result.sum() + observation_rate
    return result


"""
This submodule implements RegularizedOptimization.R in Python.

It contains functions to learn an cMHN on the *full state-space* for a given data distribution and implements the L1 regularization.
"""
# author(s): Stefan Vocht

import numpy as np
from scipy.optimize import minimize

from . import Likelihood
from . import ModelConstruction

from typing import Callable


def L1(theta: np.ndarray, eps: float = 1e-05) -> float:
    """
    Computes the L1 penalty for the given theta matrix.

    Args:
        theta (np.ndarray): The theta matrix representing the cMHN.
        eps (float, optional): A small epsilon value, mainly used for the derivative. Default is 1e-05.

    Returns:
        float: The L1 penalty for the given theta matrix.
    """
    theta_ = theta.copy()
    np.fill_diagonal(theta_, 0)
    return np.sum(np.sqrt(theta_**2 + eps))


def L1_(theta: np.ndarray, eps: float = 1e-05) -> np.ndarray:
    """
    Computes the derivative of the L1 penalty for the given theta matrix.

    Args:
        theta (np.ndarray): The theta matrix representing the cMHN.
        eps (float, optional): A small epsilon value to prevent division by zero. Default is 1e-05.

    Returns:
        np.ndarray: The derivative of the L1 penalty.
    """
    theta_ = theta.copy()
    np.fill_diagonal(theta_, 0)
    return theta_ / np.sqrt(theta_**2 + eps)


def score_reg(theta: np.ndarray, pD: np.ndarray, lam: float, n: int = None, pth_space: np.ndarray = None) -> float:
    """
    Computes the score with L1 regularization for the given theta matrix.

    Args:
        theta (np.ndarray): The theta matrix representing the cMHN.
        pD (np.ndarray): Distribution provided by the training data.
        lam (float): Tuning parameter lambda for regularization.
        n (int, optional): The number of columns/rows of the theta matrix. Default is None.
        pth_space (np.ndarray, optional): Optional parameter used for communication with the gradient function to improve performance.

    Returns:
        float: The score of the current cMHN, penalized with L1 regularization.
    """
    n = n or int(np.sqrt(theta.size))
    theta = theta.reshape((n, n))

    return -(Likelihood.score(theta, pD, pth_space) - lam * L1(theta))


def grad_reg(theta: np.ndarray, pD: np.ndarray, lam: float, n: int = 0, pth_space: np.ndarray = None) -> np.ndarray:
    """
    Computes the gradient with L1 regularization for the given theta matrix.

    Args:
        theta (np.ndarray): The theta matrix representing the cMHN.
        pD (np.ndarray): Distribution provided by the training data.
        lam (float): Tuning parameter lambda for regularization.
        n (int, optional): The number of columns/rows of the theta matrix. Default is 0.
        pth_space (np.ndarray, optional): Optional parameter used to avoid recalculating pth, improving performance.

    Returns:
        np.ndarray: The gradient of the L1-regularized score.
    """
    n = n or int(np.sqrt(theta.size))
    theta_ = theta.reshape((n, n))

    return -(Likelihood.grad(theta_, pD, pth_space) - lam * L1_(theta_)).flatten()


def learn_MHN(pD: np.ndarray, init: np.ndarray = None, lam: float = 0, maxit: int = 5000,
              trace: bool = False, reltol: float = 1e-07, round_result: bool = True,
              callback: Callable = None, score_func: Callable = score_reg, jacobi: Callable = grad_reg) -> np.ndarray:
    """
    Trains a cMHN to fit a given probability distribution.

    Args:
        pD (np.ndarray): Probability distribution used to train the new model.
        init (np.ndarray, optional): Starting point for the training (initial theta). Default is None.
        lam (float, optional): Tuning parameter lambda for regularization. Default is 0.
        maxit (int, optional): Maximum number of training iterations. Default is 5000.
        trace (bool, optional): If True, convergence messages are printed (see scipy.optimize.minimize). Default is False.
        reltol (float, optional): Gradient norm must be less than reltol before successful termination. Default is 1e-07.
        round_result (bool, optional): If True, the result is rounded to two decimal places. Default is True.
        callback (Callable, optional): Function called after each iteration, must take theta as an argument. Default is None.
        score_func (Callable, optional): Score function used for training. Default is score_reg.
        jacobi (Callable, optional): Gradient function used for training. Default is grad_reg.

    Returns:
        np.ndarray: The trained model (theta matrix).
    """

    n = int(np.log2(pD.size))

    if init is None:
        init = ModelConstruction.learn_indep(pD)

    pth_space = Likelihood.generate_pTh(init)

    opt = minimize(fun=score_func, x0=init, args=(pD, lam, n, pth_space), method="BFGS", jac=jacobi,
                   options={'maxiter': maxit, 'disp': trace, 'gtol': reltol}, callback=callback)

    theta = opt.x.reshape((n, n))

    if round_result:
        theta = np.around(theta, decimals=2)

    return theta



"""
This submodule implements the Fisher information matrix for cMHN and oMHN.

It contains function for computing the FIM in Cython and for calling the
CUDA implementation.
"""

# authors: Stefan Vocht, Y. Linda Hu

from scipy.linalg.cython_blas cimport dcopy, dscal, daxpy, ddot
from libc.stdlib cimport malloc

import numpy as np
cimport numpy as np

from .PerformanceCriticalCode cimport _compute_inverse_t, _compute_inverse,\
                                      internal_kron_vec
from .ModelConstruction import q_diag
from .Likelihood import cuda_fisher

np.import_array()

def cython_fisher(
    double[:, :] theta
):
    """
    Computes the fisher information matrix for cMHN.
    The formulas are described in S. Vocht. Identifiability of
    Mutual Hazard Networks. Unpublished bachelor thesis, 2022

    Args:
        theta (np.ndarray): matrix containing the theta values

    Returns:
        np.ndarray: The Fisher information matrix for theta
    """
    cdef int i, j, s, t
    cdef int incx = 1
    cdef int incx2 = 2
    cdef int incx0 = 0
    cdef double zero = 0.
    cdef double one = 1.
    cdef double mOne = -1.
    cdef int n = theta.shape[0]
    cdef int nx = 1 << n
    cdef int nxhalf = nx // 2

    cdef double *shuffled_vec = <double *> malloc(nx * sizeof(double))
    cdef double *q_st = <double *> malloc(nx * sizeof(double))
    cdef double *dQ_st = <double *> malloc(nx * sizeof(double))
    cdef double *masked_dQ_st = <double *> malloc(nx * sizeof(double))
    cdef double *zero_mask = <double *> malloc(nx * sizeof(double))
    cdef double *dQ_ij = <double *> malloc(nx * sizeof(double))
    cdef double *swap_vec
    cdef np.ndarray[np.double_t] pth_ = np.empty(nx, dtype=np.double)

    cdef np.ndarray[np.double_t, ndim=2] fisher_mat = np.zeros((n * n, n * n))

    # get p_th
    cdef np.ndarray[np.double_t] p0 = np.zeros(nx, dtype=np.double)
    p0[0] = 1
    # get the diagonal of I-Q
    cdef np.ndarray[np.double_t, ndim=1] dg = q_diag(theta)
    daxpy(&nx, &one, &mOne, &incx0, &dg[0], &incx)
    dscal(&nx, &mOne, &dg[0], &incx)
    _compute_inverse(&theta[0,0], n, &dg[0], &p0[0], &pth_[0])

    ind_x = -1

    for s in range(n):
        # kron_vec function from the original MHN implementation
        internal_kron_vec(
            theta, s, pth_, dQ_st, 1, 0
        )
        # set up zero_mask
        dcopy(&nxhalf, &zero, &incx0, zero_mask, &incx2)
        dcopy(&nxhalf, &one, &incx0, zero_mask + 1, &incx2)

        for t in range(n):
            # the derivative of Q wrt. theta_st ( here stored in
            # masked_dQ_st) mainly depends on s
            # the difference between the theta_st is only that for
            # s != t some entries are set to zero
            # this is done here using zero_mask
            
            if s != t:
                for i in range(nx):
                    masked_dQ_st[i] = dQ_st[i] * zero_mask[i]
            else:
                dcopy(&nx, dQ_st, &incx, masked_dQ_st, &incx)

            
            # shuffle zero mask
            dcopy(&nxhalf, zero_mask, &incx, shuffled_vec, &incx2)
            dcopy(&nxhalf, zero_mask + nxhalf, &incx, shuffled_vec+1, &incx2)

            swap_vec = shuffled_vec
            shuffled_vec = zero_mask
            zero_mask = swap_vec
            # q_st described in section 4.1.1
            # compute [I-Q]^( -1) * masked_dQ_st
            _compute_inverse(&theta[0,0], n, &dg[0], masked_dQ_st, q_st)
            
            for i in range(nx):
                q_st[i] /= pth_[i]
            
            _compute_inverse_t(&theta[0,0], n, &dg[0], q_st, shuffled_vec)
            dcopy(&nx, shuffled_vec, &incx, q_st, &incx)
            ind_x += 1
            ind_y = s * n
            # fisher matrix is symmetric , compute only the upper half
            # and use the shuffle trick for more efficiency
            for i in range(s, n):

                internal_kron_vec(theta, i, pth_, dQ_ij, 1, 0)

                for k in range(nx):
                    dQ_ij[k] *= q_st[k]

                for j in range(n):
                    
                    # reshaping for shuffling
                    dcopy(&nxhalf, dQ_ij, &incx2, shuffled_vec, &incx)
                    dcopy(&nxhalf, dQ_ij+1, &incx2, shuffled_vec+nxhalf, &incx)
                    
                    # compute only the upper half
                    if ind_y >= ind_x:

                        fisher_mat[ind_x, ind_y] = ddot(
                            &nxhalf, shuffled_vec+nxhalf, &incx, &one, &incx0)
                
                        if i == j:
                            fisher_mat[ind_x, ind_y] += ddot(
                                &nxhalf, shuffled_vec, &incx, &one, &incx0)
                    
                    ind_y += 1

                    # swap dQ_ij and shuffled_vec
                    swap_vec = dQ_ij
                    dQ_ij = shuffled_vec
                    shuffled_vec = swap_vec
 
    # fill the lower triangular part
    fisher_mat += np.tril(fisher_mat.T, -1)
    return fisher_mat

def omhn_fisher(
    log_theta: np.ndarray,
    use_cuda: bool = False,
) -> np.ndarray:
    """
    Computes the Fisher information matrix for oMHN.
    
    Args:
        log_theta (np.ndarray): matrix containing the log theta values
        use_cuda (bool, optional): whether to use GPU acceleration
    
    Returns:
        np.ndarray: The Fisher information matrix for log_theta
    """

    n = log_theta.shape[1]

    # subtract observation rates from each element in each column
    cmhn_log_theta = log_theta[:-1] - log_theta[-1]
    
    # undo changes to the diagonal
    cmhn_log_theta[np.diag_indices(n)] += log_theta[-1]

    #        / theta_theta | omega_theta \
    # FIM = |--------------+--------------|
    #        \ omega_theta | omega_omega /

    theta_theta = fisher(
        cmhn_log_theta, omhn=False, use_cuda=use_cuda)

    # F_{\omega_j}{\theta_st} = \sum_{i|=j} F_{\theta_ij}{\theta_st}
    theta_theta_reshape = theta_theta.reshape(n, n, n**2, order="F")
    omega_theta = -(
        theta_theta_reshape.sum(axis=1)
        - theta_theta_reshape[np.diag_indices(n)]
    )

    omega_theta_reshape = omega_theta.reshape(n, n, n, order="F")
    i, j = np.meshgrid(np.arange(n), np.arange(n), indexing="ij")
    omega_omega = -(
        omega_theta_reshape.sum(axis=2) - omega_theta_reshape[i, j, j]
    )

    return np.block([[theta_theta, omega_theta.T], [omega_theta, omega_omega]])


def fisher(
    log_theta: np.ndarray, omhn: bool = True, use_cuda: bool = False
) -> np.ndarray:
    """
    Computes the Fisher information matrix for o/cMHN.
    
    Args:
        log_theta(np.ndarray): matrix containing the log theta values
        omhn (bool, optional): whether MHN is oMHN or cMHN
        use_cuda (bool, optional): whether to use GPU acceleration
    
    Returns:
        np.ndarray: The Fisher information matrix for log_theta
    """
    if omhn:
        return omhn_fisher(log_theta, use_cuda=use_cuda)
    else:
        if use_cuda:
            return cuda_fisher(log_theta)
        else:
            return cython_fisher(log_theta)